'''
In the frame of the FRM4DOAS project, convert files in netCDF

       Supported native formats :

                  extended ASCII format used for CINDI (format='ASCII'),
                  CCDEEV used at BIRA-IASB (format='CCDEEV')
                  MFC_BIRA used for MFC formats (binary or STD) pre-converted
                           by MFC4QDOAS BIRA converter

Author:  Caroline FAYT (Royal Belgian Institute for Space Aeronomy)
Date:  February 2018
Version:  May 2018
'''

# ------------------------------------------------------------------------------
# MODULE        netCDF_L1
# ------------------------------------------------------------------------------
# PURPOSE       In the frame of the FRM4DOAS project, convert files in netCDF
#               Supported native formats :
#
#                  extended ASCII format used for CINDI (format='ASCII'),
#                  CCDEEV used at BIRA-IASB (format='CCDEEV')
#                  MFC_BIRA used for MFC formats (binary or STD) pre-converted
#                           by MFC4QDOAS BIRA converter
# ------------------------------------------------------------------------------
# AUTHOR        Caroline FAYT
# COPYRIGHT     Royal Belgian Institute for Space Aeronomy
# CREATION DATE February 2018
#
# MODIFIED      28 May 2018 : account for changes discussed after the demonstration meeting (8 of May 2018)
#
# ------------------------------------------------------------------------------

import configparser;
import pkgutil;

pandasFlag=1 if pkgutil.find_loader("pandas") else 0

if pandasFlag:
   import pandas as pd;      # for AERONET

from tools.netCDF_tools import *
from tools.ASCII_ExtendedFormat import *
from tools.uvvis_gb_CCD import *
from tools.MFC_Format import *

netCDF_fillValue=float('NaN');
netCDF_measurementTypes=dict([('offaxis',QDOAS_MeasType_offaxis),         \
                              ('direct sun',QDOAS_MeasType_ds),           \
                              ('zenith',QDOAS_MeasType_zenith),           \
                              ('almucantar',QDOAS_MeasType_almucantar),   \
                              ('horizon',QDOAS_MeasType_horizon),         \
                              ('direct moon',QDOAS_MeasType_moon)])

class netCDF_L1_attributes:
  '''
    Definition of attributes for netCDF L1 radiance files

    Instrument identification

       instr_number               instrument number
       instr_channel              channel number : 1, 2, 3 (according to metadata)
       instr_type                 maxdoas, zenith, mobile...

    Data_submitter

       institution                the name of the institute
       pi_name                    the name of the pi
       pi_email                   the email of the pi
       do_name                    the name of the data creator
       do_email                   the email of the data creator
       ds_name                    the name of the data submitter
       ds_email                   the email of the data submitter

    Instrument location

       station_name               the name of the station (should come from a table)
       instr_latitude             the latitude of the instrument
       instr_longitude            the longitude of the instrument
       instr_altitude_asl         the altitude of the instrument above sea level
       instr_altitude_agl         the altitude of the instrument above ground level

       prefix                     here, the project financier
       project                    could be changed in the future
  '''
  def __init__(self,instr_number='', \
                    instr_channel='', \
                    instr_type='maxdoas',\
                    institution='', \
                    pi_name='', \
                    pi_email='', \
                    do_name='', \
                    do_email='', \
                    ds_name='', \
                    ds_email='', \
                    campaign_name='', \
                    station_name='', \
                    instr_latitude='0.', \
                    instr_longitude='0.', \
                    instr_altitude_asl='0.', \
                    instr_altitude_agl='0.'):
       '''
       constructor for netCDF_L1_attributes: initialize the class variables

       '''

       # Instrument

       self.instr_number=instr_number;               # instrument number
       self.instr_channel=instr_channel;             # channel number : 1, 2, 3 (according to metadata)
       self.instr_type=instr_type;                   # maxdoas, zenith, mobile...

       # data_submitter

       self.institution=institution;                 # the name of the institute
       self.pi_name=pi_name;                         # the name of the pi
       self.pi_email=pi_email;                       # the email of the pi
       self.do_name=do_name;                         # the name of the data creator
       self.do_email=do_email;                       # the email of the data creator
       self.ds_name=ds_name;                         # the name of the data submitter
       self.ds_email=ds_email;                       # the email of the data submitter

       # instrument location

       self.station_name=station_name;               # the name of the station (should come from a table)
       self.instr_latitude=instr_latitude;           # the latitude of the instrument
       self.instr_longitude=instr_longitude;         # the longitude of the instrument
       self.instr_altitude_asl=instr_altitude_asl;   # the altitude of the instrument above sea level
       self.instr_altitude_agl=instr_altitude_agl;   # the altitude of the instrument above ground level

       self.prefix='ESA';                            # here, the project financier
       self.project='FRM4DOAS';                      # could be changed in the future
       self.campaign_name=campaign_name;

  def LoadFromIni(self,configFile,expectedSections=[]):
     """
     Initialize netCDF_L1_attributes class from options found in a config file (Windows ini style)

     Parameters
     ----------

     configFile       : the name of the ini configuration file                 \n
     expectedSections : optional, the list of expected sections                \n
     """
     config=configparser.ConfigParser()
     functionName='LoadFromIni';

     if not os.path.isfile(configFile):
       print('['+functionName+'] '+configFile+' not found !!!');
     else:

       try:

           config.read(configFile);
           configSections=config.sections();

           if not len(configSections):
             print('['+functionName+'] no sections found in '+configFile);
           else:
             sectionsToFind=expectedSections if len(expectedSections) else configSections;

             for section in sectionsToFind :

               for item in config[section]:
                 setattr(self,item,config[section][item]);

       except configparser.Error :
           print('['+functionName+'] '+configFile+' can not be loaded !!!');

       # TODO : check the different values are not empty and valid
       #
       #        instrument number, channel
       #        instrument type from metadata
       #        names and emails of pi,do,ds from a table
       #        station name from a table
       #        latitude,longitude within a radius of x km
       #        values of the altitudes w.r.t. the altitude of the station


class netCDF_L1_geodata :
  '''
  Definition of RADIANCE.GEODATA group in netCDF L1 radiance files

       See the description of the format in the appropriate documents
  '''
  def __init__(self,detsize,nrecords):
     '''
     constructor for netCDF_L1_geodata: initialize the class variables

     '''
     if (detsize>0) or (nrecords>0):

         self.viewingElevation=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.viewingAzimuth=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.solarZenithAngle=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.solarAzimuthAngle=ones((nrecords,),dtype=float32)*netCDF_fillValue;

class netCDF_L1_observations :
  '''
  Definition of RADIANCE.OBSERVATIONS group in netCDF L1 radiance files

       See the description of the format in the appropriate documents
  '''
  def __init__(self,detsize,nrecords):
     '''
     constructor for netCDF_L1_observations: initialize the class variables and arrays

     '''
     if (detsize>0) or (nrecords>0):

         self.wavelength=ones((nrecords,detsize),dtype=float32)*netCDF_fillValue;
         self.spectra=ones((nrecords,detsize),dtype=float32)*netCDF_fillValue;
         self.errors=ones((nrecords,detsize),dtype=float32)*netCDF_fillValue;
         self.pixelQualityFlags=ones((nrecords,detsize),dtype=int16);

         self.utTime=zeros((nrecords,7),dtype=int16);
         self.utStart=zeros((nrecords,7),dtype=int16);
         self.utEnd=zeros((nrecords,7),dtype=int16);
         self.exposureTime=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.measureType=ones((nrecords,),dtype=int16)*(-1);
         self.nacc=ones((nrecords,),dtype=int16)*(-1);

         self.totalAcqTime=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.totalMeasTime=ones((nrecords,),dtype=float32)*netCDF_fillValue;

class netCDF_L1_radiance :
  '''
  Definition of RADIANCE group in netCDF L1 radiance files

       This group is divided in two subgroups (GEODATA and OBSERVATIONS).  See the description of the format in the appropriate documents
  '''
  def __init__(self,detsize,nrecords):
     '''
     constructor for netCDF_L1_radiance: initialize the class variables

     '''
     if (detsize>0) or (nrecords>0):

         self.geodata=netCDF_L1_geodata(detsize,nrecords);
         self.observations=netCDF_L1_observations(detsize,nrecords);

class netCDF_L1_records :
  '''
  Definition of the struecture of records to read/save in netCDF L1 radiance files

       See the description of the format in the appropriate documents
  '''
  def __init__(self,detsize,nrecords):
     '''
     constructor for netCDF_L1_records: initialize the class variables

     '''

     self.detsize=detsize;
     self.nrecords=nrecords;
     self.datesize=7;

     if (detsize>0) or (nrecords>0):

         self.radiance=netCDF_L1_radiance(detsize,nrecords);

class netCDF_L1_tp :
  '''
  Definition of ANCILLARY.METEOROLOGICAL_DATA temperature and pressure profiles groups in netCDF L1 radiance files

       See the description of the format in the appropriate documents
  '''
  def __init__(self,tpFile):
     '''
     constructor for netCDF_L1_tp: initialize the class variables.  Information on temperature and pressure profiles are retrieved from a tpFile

     '''

     functionName='netCDF_L1_tp'

     self.altitude_level=[];
     self.meteo_time=[];
     self.pressure=[];
     self.temperature=[];

     self.nlevels=0;
     self.ntime=0;

     if len(tpFile):

       if not os.path.isfile(tpFile):
         print('['+functionName+'] '+tpFile+' not found !!!');
       else:

         tp=loadtxt(tpFile);

         self.nlevels=size(tp,0);
         self.ntime=5;   # could be changed later

         self.meteo_time[:]=array([0.,6.,12.,18.,24.]);

         self.altitude_level=tp[:,0];
         self.pressure=zeros((self.nlevels,self.ntime));
         self.temperature=zeros((self.nlevels,self.ntime));

         for i in arange(0,self.ntime):

           self.pressure[:,i]=tp[:,2];
           self.temperature[:,i]=tp[:,3];

class netCDF_L1_cloud :
  '''
  Definition of ANCILLARY.METEOROLOGICAL_DATA.CLOUND_INFORMATION group in netCDF L1 radiance files

       See the description of the format in the appropriate documents
  '''
  def __init__(self,cloudFile):
     '''
     constructor for netCDF_L1_cloud: initialize the class variables.  Information on cloud are retrieved from a cloudFile

     '''
     functionName='netCDF_L1_cloud'

     self.cloud_time=[];
     self.cloud_coverage=[];
     self.cloud_height=[];
     self.cloud_size=0;

     if len(cloudFile):

       if not os.path.isfile(cloudFile):
         print('['+functionName+'] '+cloudFile+' not found !!!');
       else:

         cloud=loadtxt(cloudFile);
         cloud_size=shape(cloud);

         if (ndim(cloud)!=2) or (cloud_size[1]<3):
           print('['+functionName+'] '+cloudFile+' should contain at least 3 columns !!!');
         elif cloud_size[0]>0:

           self.cloud_size=cloud_size[0];   # could be changed later

           self.cloud_time[:]=cloud[:,0];
           self.cloud_coverage[:]=cloud[:,1];
           self.cloud_height[:]=cloud[:,2];

# +++ AERONET +++ This function should move to doas_lib.py later
#                 GB_Batch_... : separation between analysis (centOS) and web (kronos)

def read_aeronet(filename,column_index='Day_of_Year(Fraction)',skiprows=6):
    """
    Read a given AERONET AOT data file, and return it as a dataframe.

    This Return a DataFrame containing the AERONET data, with the index
    set to the timestamp of the AERONET observations. Rows or columns
    consisting entirely of missing data are removed. All other columns
    are left as-is.

    This function has been download from this URL :

    http://blog.rtwilson.com/reading-aeronet-data-in-pandas-a-simple-helper-function/

    Parameters
    ----------

    filename           : the name of the AERONET AOT data file                               \n
    column_index       : the title of the column for time (by default, Day_of_Year(Fraction) \n
    skiprows           : the number of lines of the header to skip                           \n

    Returns
    -------

    an : the aeronet data frame
    """
    an=[];

    if pandasFlag:

       dateparse = lambda x: pd.datetime.strptime(x, "%d:%m:%Y %H:%M:%S")
       aeronet = pd.read_csv(filename, skiprows=skiprows, na_values=['N/A'],
                             parse_dates={'times':[0,1]},
                             date_parser=dateparse)

       aeronet = aeronet.set_index(column_index)
       # del aeronet['Julian_Day']

       # Drop any rows that are all NaN and any cols that are all NaN
       # & then sort by the index
       an = (aeronet.dropna(axis=1, how='all')
                    .dropna(axis=0, how='all')
                    .rename(columns={'Last_Processing_Date(dd/mm/yyyy)': 'Last_Processing_Date'})
                    .sort_index())

    return an

class netCDF_L1_aod :
  '''
  Definition of ANCILLARY.AEROSOL_DATA group in netCDF L1 radiance files

       See the description of the format in the appropriate documents
  '''
  def __init__(self,aodFile):
     '''
     constructor for netCDF_L1_aod initialize the class variables.  The information on aerosols are retrieved from aodFile

     '''

     functionName="netCDF_L1_aod"

     self.asymmetry=[];   # asymmetry factor
     self.alpha=[];       # angstrom coefficient
     self.albedo=[];      # single scattering albedo
     self.aod=[];
     self.wavelength=[];
     self.time=[];

     self.nwve=0;
     self.ntime=0;

     if len(aodFile):

       if not os.path.isfile(aodFile):
         print('['+functionName+'] '+aodFile+' not found !!!');
       elif not pandasFlag:
         print('['+functionName+'] module pandas not found on this machine');
       else:

         data=read_aeronet(aodFile,skiprows=6,column_index='Day_of_Year(Fraction)');

         header=list(data.columns.values);

         # Get wavelengths

         for item in header:
             if re.sub(r'^AOD_([0-9]*)nm$', r'\1',item)!=item and max(pd.Series.as_matrix(data[item])>0):
                self.wavelength.append(float(re.sub(r'^AOD_([0-9]*)nm$', r'\1',item)));

         if (len(self.wavelength)):

             self.wavelength.sort();
             dd=asarray(data.index);
             self.time=(dd-floor(dd))*24.;
             self.nwve=len(self.wavelength);
             self.ntime=len(self.time);
             self.aod=zeros((self.ntime,self.nwve))
             self.asymmetry=ones((self.ntime,self.nwve))*nan
             self.alpha=ones(self.ntime)
             self.albedo=ones((self.ntime,self.nwve))*nan

             for i,wve in enumerate(self.wavelength):
                self.aod[:,i]=pd.Series.as_matrix(data[sprintf('AOD_%gnm',wve)]);
                j=[j for j,item in enumerate(self.aod[:,i]) if item<-990.];
                if len(j)>0:
                   self.aod[j,i]=NaN;

def netCDF_L1_LoadfromASCII(ascFile):
    """
    Load Level-1 radiance spectra from a file in ASCII extended format

    Parameters
    ----------

    ascFile : the name of the file to read

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, ReadExtendedAsc (from ASCII_ExtendedFormat)
    """

    functionName="netCDF_LoadL1_fromASCII"
    L1=netCDF_L1_records(0,0);

    if not os.path.isfile(ascFile):
      print('['+functionName+'] '+ascFile+' not found !!!');
    else:

      data=ReadExtendedAsc(ascFile);

      nrecords=len(data);
      new_nrecords=nrecords;
      detsize=0 if nrecords==0 else len(data[0].spectrum);

      if (detsize==0) or (nrecords==0):
        print('['+functionName+'] '+ascFile+' is empty !!!');
      else:

        for i in arange(0,nrecords):
           if data[i].wavelength[0]<250. or data[i].wavelength[0]>850.:
             new_nrecords=new_nrecords-1;

        if (new_nrecords!=nrecords):
           print('Number of rejected spectra : '+str(nrecords-new_nrecords));

        L1=netCDF_L1_records(detsize,new_nrecords);
        j=0;

        for i in arange(0,nrecords):

          if data[i].wavelength[0]>250. and data[i].wavelength[0]<850.:

             header=data[i].header;

             ut_day, ut_month, ut_year = header['date'].split('/')
             uts_hour, uts_min, uts_sec = header['utc start time'].split(':')
             ute_hour, ute_min, ute_sec = header['utc end time'].split(':')
             ut_hour, ut_min, ut_sec = header['utc time'].split(':')

             uts_ms=str(int((float(uts_sec)-int(float(uts_sec)))*1000))
             uts_sec=str(int(float(uts_sec)))
             ute_ms=str(int((float(ute_sec)-int(float(ute_sec)))*1000))
             ute_sec=str(int(float(ute_sec)))
             ut_ms=str(int((float(ut_sec)-int(float(ut_sec)))*1000))
             ut_sec=str(int(float(ut_sec)))

             L1.radiance.observations.utStart[j,:] = array([ut_year, ut_month, ut_day, uts_hour, uts_min, uts_sec, uts_ms], dtype=int16)
             L1.radiance.observations.utEnd[j,:] = array([ut_year, ut_month, ut_day, ute_hour, ute_min, ute_sec, ute_ms], dtype=int16)
             L1.radiance.observations.utTime[j,:] = array([ut_year, ut_month, ut_day, ut_hour, ut_min, ut_sec, ut_ms], dtype=int16)

             L1.radiance.geodata.viewingElevation[j]=header['viewing elevation angle'];
             L1.radiance.geodata.viewingAzimuth[j]=header['viewing azimuth angle'];

             if 'solar zenith angle' in header and header['solar zenith angle']>0. and header['solar zenith angle']<100. :
                L1.radiance.geodata.solarZenithAngle[j]=header['solar zenith angle'];
             if 'solar azimuth angle' in header:
                L1.radiance.geodata.solarAzimuthAngle[j]=header['solar azimuth angle'];

             L1.radiance.observations.exposureTime[j]=header['exposure time (sec)'];
             L1.radiance.observations.nacc[j]=-1;  # -1 if not available; to change if available

             L1.radiance.observations.measureType[j]=header['measurement type'];

             if (L1.radiance.geodata.viewingElevation[j]>80.) and (header['measurement type']==QDOAS_MeasType_offaxis):
                L1.radiance.observations.measureType[j]=QDOAS_MeasType_zenith;

             L1.radiance.observations.totalAcqTime[j]=header['total acquisition time (sec)'];
             L1.radiance.observations.totalMeasTime[j]=header['total measurement time'];

             L1.radiance.observations.wavelength[j,:]=data[i].wavelength;
             L1.radiance.observations.spectra[j,:]=data[i].spectrum;

             j=j+1;

    return L1;

def readSwingFile(fil):

    IGNORED_BYTES = 153

    dtype = np.dtype([

    ("year", np.int16),
    ("day", np.int8),
    ("month", np.int8),

    ("min", np.int8),
    ("hour", np.int8),
    ("hund", np.int8),
    ("sec", np.int8),

    ("avgdSpectra", np.int32),
    ("totalTime", np.int32),
    ("nrejMeas", np.int32),
    ("naccMeas", np.int32),

    ("longitude",np.float32),
    ("latitude",np.float32),
    ("altitude",np.float32),

    ("exposureTime",np.float32),

    ("minGps", np.int8),
    ("hourGps", np.int8),
    ("hundGps", np.int8),
    ("secGps", np.int8),

    ("floatflag",np.int8),
    ("position",np.int8),
    ("receivedPosition",np.int8),
    # spectrum = fread(fp,2048,'double');

    ("ignoredBytes",np.int8,IGNORED_BYTES),

    ("spectrum",np.float64,2048)

    ])

    with open(fil, "rb") as f:
        # Read the whole file at once
        data = np.fromfile(f, dtype=dtype)

    return data

def netCDF_L1_LoadfromSwing(speFile,clbFile='',commentChar=';'):

    """
    L1=netCDF_L1_LoadfromSwing(dscFile,[clbFile])

    Load Level-1 radiance spectra from a file in Swing format
    This format is used at BIRA-IASB

    Parameters
    ----------

    dscFile : the name of the file to read
    clbFile : optional but recommended, the wavelength calibration to associate to the file

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, readSwingFile (from uvvis_gb_CCD)
    """

    functionName="netCDF_L1_LoadfromSwing"
    L1=netCDF_L1_records(0,0);

    clb=[];
    if len(clbFile):
       clb=netCDF_L1_LoadCalib(clbFile,commentChar=commentChar);

    if not os.path.isfile(speFile):
      print('['+functionName+'] '+dscFile+' not found !!!');
    else :

      data=readSwingFile(speFile);

      nrecords=len(data);
      detsize=2048;

      if nrecords==0:
        print('['+functionName+'] '+dscFile+' is empty !!!');
      elif (len(clb)>0) and (len(clb)!=detsize):
        print('['+functionName+'] calibration '+clbFile+' should have the same number of points as spectra !!!');
      else:

        L1=netCDF_L1_records(detsize,nrecords);

        for i in arange(0,nrecords):

          L1.radiance.observations.utTime[i,:]=array([data["year"][i],data["month"][i],data["day"][i],data["hour"][i],data["min"][i],data["sec"][i],0],dtype=int16)

          L1.radiance.observations.utStart[i,:]=array([data["year"][i],data["month"][i],data["day"][i],data["hour"][i],data["min"][i],data["sec"][i],0],dtype=int16)

          L1.radiance.observations.utEnd[i,:]=array([data["year"][i],data["month"][i],data["day"][i],data["hour"][i],data["min"][i],data["sec"][i],0],dtype=int16)

          L1.radiance.geodata.viewingElevation[i]=data["position"][i]

          L1.radiance.geodata.viewingAzimuth[i]=0.

          L1.radiance.geodata.solarZenithAngle[i]=90.

          L1.radiance.geodata.solarAzimuthAngle[i]=0.

          L1.radiance.observations.exposureTime[i]=data['exposureTime'][i]

          L1.radiance.observations.nacc[i]=data['avgdSpectra'][i]

          L1.radiance.observations.measureType[i]=CCD_MT_offaxis if data["position"][i]<80. else CCD_MT_zenith;

          L1.radiance.observations.totalAcqTime[i]=data['exposureTime'][i]*data["avgdSpectra"][i]/1000.

          L1.radiance.observations.totalMeasTime[i]=0.

          if len(clb):
            L1.radiance.observations.wavelength[i,:]=clb;

          L1.radiance.observations.spectra[i,:]=data['spectrum'][i]

    return L1;

def netCDF_L1_LoadfromCCDEEV(dscFile,clbFile=''):
    """
    Load Level-1 radiance spectra from a file in CCD-EEV format

    This format is used at BIRA-IASB
    For FRM4DOAS, only DSC files (dark current and straylight corrected spectra should be used)

    Parameters
    ----------

    dscFile : the name of the file to read
    clbFile : optional but recommended, the wavelength calibration to associate to the file

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, CCD_ReadSpe (from uvvis_gb_CCD)
    """

    functionName="netCDF_L1_LoadfromCCDEEV"
    L1=netCDF_L1_records(0,0);

    clb=[];
    if len(clbFile):
       clb=netCDF_L1_LoadCalib(clbFile,commentChar=';');

    if not os.path.isfile(dscFile):
      print('['+functionName+'] '+dscFile+' not found !!!');
    else :

      data=CCD_ReadSpe(dscFile,0,0,0,0,0);

      nrecords=len(data);
      detsize=0 if nrecords==0 else len(data[0].spectrum);

      if (detsize==0) or (nrecords==0):
        print('['+functionName+'] '+dscFile+' is empty !!!');
      elif (len(clb)>0) and (len(clb)!=detsize):
        print('['+functionName+'] calibration '+clbFile+' should have the same number of points as spectra !!!');
      else:

        L1=netCDF_L1_records(detsize,nrecords);

        for i in arange(0,nrecords):

          header=data[i].header;
          measType=header['measureType'];

          if (measType==CCD_MT_offaxis) or (measType==CCD_MT_azimuth) :
             measType=CCD_MT_offaxis;
          elif (measType==CCD_MT_ds) or (measType==CCD_MT_almucantar) or (measType==CCD_MT_horizon) :
             measType=measType;
          else:
             measType=CCD_MT_zenith;

          trackerFlag=(header['brusagElevation']!=header['brusagAzimuth']) or ((header['brusagElevation']!=-1.) and (header['brusagAzimuth']!=-1.));

          L1.radiance.observations.utTime[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['ti_hour'],header['ti_min'],header['ti_sec'],0],dtype=int16);
          L1.radiance.observations.utStart[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['start_ti_hour'],header['start_ti_min'],header['start_ti_sec'],0],dtype=int16);
          L1.radiance.observations.utEnd[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['end_ti_hour'],header['end_ti_min'],header['end_ti_sec'],0],dtype=int16);

          # TO DO : check utStart and utEnd !!!

          L1.radiance.geodata.viewingElevation[i]=90. if trackerFlag==False else floor(header['targetElevation']+0.5);
          L1.radiance.geodata.viewingAzimuth[i]=header['targetAzimuth']; # ??? should be better for BXLS if trackerFlag==False else floor(header['brusagAzimuth']+0.5);
          L1.radiance.geodata.solarZenithAngle[i]=header['Zm'];
          L1.radiance.geodata.solarAzimuthAngle[i]=header['Azimuth'];

          # print(L1.radiance.geodata.viewingElevation[i],L1.radiance.geodata.viewingAzimuth[i])

          L1.radiance.observations.exposureTime[i]=header['exposureTime'];
          L1.radiance.observations.nacc[i]=header['nacc'];
          L1.radiance.observations.measureType[i]=measType;
          L1.radiance.observations.totalAcqTime[i]=float(header['nacc']*header['exposureTime']);
          L1.radiance.observations.totalMeasTime[i]=float((header['end_ti_hour']*3600.+header['end_ti_min']*60.+header['end_ti_sec'])-   \
                                             (header['start_ti_hour']*3600.+header['start_ti_min']*60.+header['start_ti_sec']));

          if len(clb):
            L1.radiance.observations.wavelength[i,:]=clb;
          L1.radiance.observations.spectra[i,:]=reshape(data[i].spectrum*L1.radiance.observations.nacc[i],(detsize,));

    return L1;

def netCDF_L1_LoadfromMFC_BIRA(mfcFile,clbFile='',commentChar='#'):
    """
    Load Level-1 radiance spectra from a file in MFC BIRA-IASB format

    The format is proper to BIRA-IASB and is generated by a converter (MFC4QDOAS_Converter) from original MFC files (binary or std)
    produced by the acquisition S/W DOASIS.  In the BIRA-IASB, all files what ever the spectrum type (dark current, offset and observations)
    are saved in daily files.  The function MFC_BIRA_ReadSpe (such as QDOAS) reads and average dark current and offset and apply
    the correction to observations.

    Parameters
    ----------

    mfcFile     : the name of the file to read (MFC BIRA-IASB)
    clbFile     : optional but recommended, the wavelength calibration to associate to the file
    commentChar : the character for comments used in clbFile

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, MFC_BIRA_ReadSpe (from MFC_format)
    """

    functionName="netCDF_L1_LoadfromMFC_BIRA"
    L1=netCDF_L1_records(0,0);
    clb=[];
    if len(clbFile):
       clb=netCDF_L1_LoadCalib(clbFile,commentChar=commentChar);

    if not os.path.isfile(mfcFile):
      print('['+functionName+'] '+mfcFile+' not found !!!');
    else :

      data=MFC_BIRA_ReadSpe(mfcFile,1);

      nrecords=len(data);
      detsize=0 if nrecords==0 else len(data[0].spectrumCorrected);

      if (detsize==0) or (nrecords==0):
        print('['+functionName+'] '+mfcFile+' is empty !!!');
      elif (len(clb)>0) and (len(clb)!=detsize):
        print('['+functionName+'] calibration '+clbFile+' should have the same number of points as spectra !!!');
      else:

        L1=netCDF_L1_records(detsize,nrecords);

        for i in arange(0,nrecords):

          header=data[i].header;
          measType=header['measurementType'];

          if (measType==CCD_MT_offaxis) or (measType==CCD_MT_azimuth) :
             measType=CCD_MT_offaxis;
          elif (measType==CCD_MT_ds) or (measType==CCD_MT_almucantar) or (measType==CCD_MT_horizon) :
             measType=measType;
          else:
             measType=CCD_MT_zenith;

          L1.radiance.observations.utStart[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['start_ti_hour'],header['start_ti_min'],header['start_ti_sec'],0],dtype=int16);
          L1.radiance.observations.utEnd[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['end_ti_hour'],header['end_ti_min'],header['end_ti_sec'],0],dtype=int16);

          dt=datevec((datenum(header['da_year'],header['da_month'],header['da_day'],header['start_ti_hour'],header['start_ti_min'],header['start_ti_sec'])+ \
                     datenum(header['da_year'],header['da_month'],header['da_day'],header['end_ti_hour'],header['end_ti_min'],header['end_ti_sec']))*0.5);

          L1.radiance.observations.utTime[i,0:6]=array(dt,dtype=int16);
          L1.radiance.observations.utTime[i,6]=0;

          L1.radiance.geodata.viewingElevation[i]=header['elevationAngle'];
          L1.radiance.geodata.viewingAzimuth[i]=header['azimuthAngle'];
          L1.radiance.geodata.solarZenithAngle[i]=fillValue;    # todo : should be added by the MFC4QDOAS converter
          L1.radiance.geodata.solarAzimuthAngle[i]=fillValue;   # todo : should be added by the MFC4QDOAS converter header['Azimuth'];

          L1.radiance.observations.exposureTime[i]=header['exposureTime'];
          L1.radiance.observations.nacc[i]=header['scansNumber'];
          L1.radiance.observations.measureType[i]=measType;
          L1.radiance.observations.totalAcqTime[i]=float(header['scansNumber']*header['exposureTime']);
          L1.radiance.observations.totalMeasTime[i]=float((header['end_ti_hour']*3600.+header['end_ti_min']*60.+header['end_ti_sec'])-   \
                                             (header['start_ti_hour']*3600.+header['start_ti_min']*60.+header['start_ti_sec']));

          if len(clb):
            L1.radiance.observations.wavelength[i,:]=clb;
          L1.radiance.observations.spectra[i,:]=reshape(data[i].spectrumCorrected,(detsize,));

    return L1;

def netCDF_L1_LoadfromNetCDF(ncFile):
    """
    Load Level-1 radiance spectra from a file in netCDF format

    Parameters
    ----------

    ncFile : the name of the file to read

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class
    """

    functionName="netCDF_L1_LoadfromNetCDF"

    if not os.path.isfile(ncFile):
      print('['+functionName+'] '+ncFile+' not found !!!');
    else:

      try:

        nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');

        nrecords=nc.dimensions['number_of_records'].size;
        detsize=nc.dimensions['detector_size'].size;

        if nrecords>0:

             L1=netCDF_L1_records(detsize,nrecords);

             # radiance.observations

             L1.radiance.observations.utTime=asarray(nc.groups['radiance'].groups['observations'].variables['datetime'])
             L1.radiance.observations.utStart=asarray(nc.groups['radiance'].groups['observations'].variables['datetime_start'])
             L1.radiance.observations.utEnd=asarray(nc.groups['radiance'].groups['observations'].variables['datetime_end'])

             L1.radiance.geodata.viewingElevation=asarray(nc.groups['radiance'].groups['geodata'].variables['elevation_viewing_angle'])
             L1.radiance.geodata.viewingAzimuth=asarray(nc.groups['radiance'].groups['geodata'].variables['azimuth_viewing_angle'])
             L1.radiance.geodata.solarZenithAngle=asarray(nc.groups['radiance'].groups['geodata'].variables['solar_zenith_angle'])
             L1.radiance.geodata.solarAzimuthAngle=asarray(nc.groups['radiance'].groups['geodata'].variables['solar_azimuth_angle'])

             L1.radiance.observations.exposureTime=asarray(nc.groups['radiance'].groups['observations'].variables['exposure_time'])
             L1.radiance.observations.nacc=asarray(nc.groups['radiance'].groups['observations'].variables['number_of_coadded_spectra'])
             L1.radiance.observations.measureType=asarray(nc.groups['radiance'].groups['observations'].variables['measurement_type'])

             L1.radiance.observations.totalAcqTime=asarray(nc.groups['radiance'].groups['observations'].variables['total_acquisition_time'])
             L1.radiance.observations.totalMeasTime=asarray(nc.groups['radiance'].groups['observations'].variables['total_measurement_time'])

             # Create measurements variables

             L1.radiance.observations.wavelength=asarray(nc.groups['measurements'].variables['wavelengths'])
             L1.radiance.observations.spectra=asarray(nc.groups['measurements'].variables['radiances'])
             # L1.radiance.observations.errors=asarray(nc.groups['measurements'].variables['errors'])
             L1.radiance.observations.pixelQualityFlags=asarray(nc.groups['measurements'].variables['radiances_quality_flags'])

        nc.close();

      except BaseException as error:
        print('['+functionName+'] returned fatal error : {}'.format(error))
        return;

    return L1;

def netCDF_L1_LoadCalib(clbFile,commentChar='#'):
    """
    Load the calibration file to associate to L1 radiance spectra

    It should be a ASCII file.  The character to use for comments should be provided to loadtxt function if it is not the python one (#)

    Parameters
    ----------

    clbFile     : the name of the calibration file
    commentChar : optional, comment character to pass to loadtxt

    Returns
    -------

    the wavelength calibration grid
    """

    # to do : check the wavelength coverage ?

    clb=[];

    functionName="netCDF_L1_LoadCalib";
    if len(clbFile)>0:

       if not os.path.isfile(clbFile):
          print('['+functionName+'] '+clbFile+' not found !!!');
       else:

         clb=loadtxt(clbFile,comments=commentChar);
         if (clb.ndim>1):
           clb=clb[:,0];

    return clb;

def netCDF_L1_LoadRef(refFile,commentChar='#'):
    """
    Load the reference spectrum that will be used for the wavelength calibration

    It should be a two-columns ASCII file.  The character to use for comments should be provided to loadtxt function if it is not the python one (#)

    Parameters
    ----------

    refFile     : the name of the reference file
    commentChar : optional, comment character to pass to loadtxt

    Returns
    -------

    the reference calibration and spectrum
    """

    # to do : check the wavelength coverage and the S/N ratio of the reference ?

    ref=[];

    functionName="netCDF_L1_LoadRef";
    if len(refFile)>0:

       if not os.path.isfile(refFile):
          print('['+functionName+'] '+refFile+' not found !!!');
       else:

         ref=loadtxt(refFile,comments=commentChar);
         if (ref.ndim<2):
            print('['+functionName+'] '+refFile+' should contain at least two columns');

    return ref;

def netCDF_L1_LoadSlitFunction(slfFile,wavelength=0.,commentChar='#'):
    """
    Load the slit funtion that will be used for the wavelength calibration

    It should be a multiple columns ASCII file.  The character to use for comments should be provided to loadtxt function if it is not the python one (#)
    For tables, the first line should contain the different wavelengths at which the slit function have been measured.
    If only one slit function is given, the wavelength should be provided

    Parameters
    ----------

    slfFile     : the name of the slit function file
    wavelength  : if only one line shape is given, it is the wavelength at wich the slit function has been measured
    commentChar : optional, comment character to pass to loadtxt

    Returns
    -------

    the table of slit functions
    """

    # to do : check the wavelength calibration (should be relative)
    #         check the slit functions ??? estimation of the FWHM ?
    #         check wavelengths at which the slit function has been measured
    #         allow wavelength dependent slit functions (wavelength dependend gaussian file produced by QDOAS ?)

    slf=[];
    slitX=[];
    slitY=[];
    slit=[];

    functionName="netCDF_L1_LoadSlitFunction";
    if len(slfFile)>0:

       if not os.path.isfile(slfFile):
          print('['+functionName+'] '+slfFile+' not found !!!');
       else:

         slf=loadtxt(slfFile,comments=commentChar);
         if (slf.ndim<2):
            print('['+functionName+'] '+slfFile+' should contain at least two columns');
         elif size(slf,1)==2 and wavelength<300.:
           print('['+functionName+'] wavelength at which the slit function has been measured not specified');
         elif size(slf,1)==2:

             slitX=slf[:,0];
             slitY=array([wavelength]);
             slit=slf[:,1];

         else:  # table of slit functions

             slitX=slf[1:size(slf,0),0];
             slitY=slf[0,1:size(slf,1)];
             slit=slf[1:size(slf,0),1:size(slf,1)]; # measured slit functions

    return (slitX,slitY,slit);

def netCDF_L1_BuildFileName(ncPath,                          \
                            startDatetime,endDatetime,       \
                            prefix="ESA",                    \
                            project="FRM4DOAS",              \
                            institution="BIRA.IASB",         \
                            station_name="HARESTUA",         \
                            campaign_name="",                \
                            instr_number="004",              \
                            instr_channel="1",               \
                            file_type="L1",                  \
                            file_version="001"):
    """
    Build the name of the L1 radiance file according to conventions predefined by the current project.

    - is the field separator

    Parameters
    ----------

    ncPath            : the path in which to create the file name
    startDatetime     : the date and time of the first measurement (should be a yyyy,mm,dd,hh,mm,ss,ms array)
    endDatetime       : the date and time of the last measurement (should be a yyyy,mm,dd,hh,mm,ss,ms array)
    prefix            : optional, the prefix of the file name (should be ESA for ESA projects)
    project           : optional, the name of the project (should be FRM4DOAS for FRM4DOAS)
    institution       : optional, the name of the institute
    station_name      : optional, the name of the station
    campaign_name     : optional, the name of the campaign
    instr_number      : optional, the instrument number (should be unique)
    instr_channel     : optional, the channel number
    file_type         : optional, the file type (should be L1 for L1 radiance files)
    file_version      : optional, the file version

    Returns
    -------

    the name of the netCDF file built with input fields
    """

    function_name="netCDF_L1_BuildFileName"
    ncFile="";

  # todo : check the format of startDatetime and endDatetime
  #        check that the institution in data base
  #        check that the station name in the data base
  #        according to the instrument number, check the consistency with instrument channel, institution and station

    if not os.path.isdir(ncPath):
      print('['+function_name+'] '+ncPath+' not a valid directory');
    elif len(prefix)==0 or len(project)==0 or len(institution)==0 or len(station_name)==0 or len(instr_number)==0 or len(instr_channel)==0 or len(file_type)==0 or len(file_version)==0:
      print('['+function_name+'] one of the input arguments is empty');
    else:

      location=campaign_name if len(campaign_name)>0 else station_name;

      d1='%d%02d%02d' % (startDatetime[0],startDatetime[1],startDatetime[2]);
      d2='%d%02d%02d' % (endDatetime[0],endDatetime[1],endDatetime[2]);
      t1='%02d%02d%02d' % (startDatetime[3],startDatetime[4],startDatetime[5]);
      t2='%02d%02d%02d' % (endDatetime[3],endDatetime[4],endDatetime[5]);

      ncFile=ncPath+"/"+prefix+"-"+project+"-"+file_type+"-"+institution+"-"+location+"-"+instr_number+"-"+instr_channel+"-"+d1+'T'+t1+"Z-"+d2+'T'+t2+"Z-fv"+file_version+".nc";

    return ncFile;

def netCDF_L1_Convert(cfgFile,inpFile,ncPath,refFile='',slfFile='',slfWavelength=0.,clbFile='', \
                      tpFile='',cloudFile='',aodFile='',surfaceAlbedo=np.nan, \
                      file_version='001',file_format='ASCII',commentChar='#', \
                      solarAzimuthCorrection=0.,viewingAzimuthCorrection=0.):
    """
    Read file in a native format and convert it into the netCDF file

    Parameters
    ----------

    cfgFile                  : ini file with some (key,value) to use as file attributes                                     \n
    inpFile                  : name of the input file (ASCII format)                                                        \n
    ncPath                   : name of the output path                                                                      \n
    refFile                  : optional, reference file for QDOAS calibration as keydata                                    \n
    slfFile                  : optional, slit function file provided by the data submitter as keydata                       \n
    slfWavelength            : optional, for 1D slit function, the wavelength at which the slit function has been measured  \n
    clbFile                  : optional, wavelength calibration file                                                        \n
    tpFile                   : optional, temperature/profile file                                                           \n
    cloudFile                : optional, cloud file (should include information on cloud coverage and heights)              \n
    aodFile                  : optional, aerosol data                                                                       \n
    surfaceAlbedo            : optional, surface albedo                                                                     \n
    file_version             : optional, the file version on three digits                                                   \n
    file_format              : the native file format (by default, ASCII but CCEDDEV and MFC-BIRA supported                 \n
    commentChar              : optional, the character to use for comment in clbFile, refFile, slfFile                      \n
    solarAzimuthCorrection   : optional, the azimuth correction to apply to solar azimuth angle (+ is applied)              \n
    viewingAzimuthCorrection : optional, the azimuth correction to apply to viewing azimuth angle (+ is applied)            \n


    TODO
    ----

    check arguments ??? for example : valid value for format

    Returns
    -------

    the name of the output netCDF file (built from information passed as input)
    """

    functionName="netCDF_L1_Convert"
    ncFile=""

    if not os.path.isfile(cfgFile):
       print('['+functionName+'] config file '+cfgFile+' not found !!!');
    elif not os.path.isfile(inpFile):
       print('['+functionName+'] input file '+inpFile+' not found !!!');
    elif not os.path.isdir(ncPath):
       print('['+functionName+']'+ncPath+' not a directory !!!');
    else:

       # Load input files

       config=netCDF_L1_attributes();  # Initialize attributes

       try:

        config.LoadFromIni(cfgFile);                                             # Load attributes from config file

        if (file_format=='CCDEEV'):
           L1_data=netCDF_L1_LoadfromCCDEEV(inpFile,clbFile);
        elif (file_format=='ASCII'):
           L1_data=netCDF_L1_LoadfromASCII(inpFile);                             # Load data from L1 in extended ASCII format
        elif (file_format=='MFC_BIRA'):
           L1_data=netCDF_L1_LoadfromMFC_BIRA(inpFile,clbFile,commentChar=commentChar);
        elif (file_format=='SWING'):
           L1_data=netCDF_L1_LoadfromSwing(inpFile,clbFile,commentChar=commentChar);
        elif (file_format=='netCDF'):
           L1_data=netCDF_L1_LoadfromNetCDF(inpFile);
        else:
           print('['+functionName+'] Unknown format '+file_format+' !!!');

        tp=netCDF_L1_tp(tpFile);                                                 # Load temperature/pressure profiles
        cloud=netCDF_L1_cloud(cloudFile);                                        # Load information on cloud
        aod=netCDF_L1_aod(aodFile);                                              # Load information on aerosols

        ref=netCDF_L1_LoadRef(refFile,commentChar=commentChar);                                          # Load reference file if any
        (slfX,slfY,slf)=netCDF_L1_LoadSlitFunction(slfFile,slfWavelength,commentChar=commentChar);       # Load slit function if any

       except BaseException as error:
        print('['+functionName+'] returned fatal error : {}'.format(error))
        return;

       nrecords=L1_data.nrecords;
       detsize=L1_data.detsize;
       refSize=shape(ref);
       slfSize=shape(slf);

       if nrecords==0 or detsize==0:
          print('['+functionName+']'+inpFile+' is empty');
       elif refSize[0]>0 and refSize[0]!=detsize:
          print('['+functionName+'] Spectra and reference do not have the same size');
       else :

          # Build filename (should be something like ESA-FRM4DOAS-L1-BIRA.IASB-HARESTUA-004-1-20171208T083023Z-20171208T151709Z-fv001.nc)

          # station_name=config.station_name, -> should be campaign or station name

          ncFile=netCDF_L1_BuildFileName(ncPath,                                \
                                         L1_data.radiance.observations.utTime[0,:],          \
                                         L1_data.radiance.observations.utTime[nrecords-1,:], \
                                         prefix=config.prefix,                  \
                                         project=config.project,                \
                                         institution=config.institution,        \
                                         station_name=config.station_name,      \
                                         campaign_name=config.campaign_name,    \
                                         instr_number=config.instr_number,      \
                                         instr_channel=config.instr_channel,    \
                                         file_type="L1",                        \
                                         file_version=file_version);

          # Open file

          try:
             nc_root=netCDF4.Dataset(ncFile, 'w', format='NETCDF4');
          except:
             print('['+functionName+'] Can not open '+ncFile+' in writing mode !!!');
             return;

          # Set attributes

          nc_root.Conventions="CF-1.6";
          nc_root.title='Level-1 data for the FRM4DOAS MAXDOAS processor';
          nc_root.institution=config.institution;
          nc_root.source='MAXDOAS calibrated radiances';

          nc_root.pi_name=config.pi_name;
          nc_root.pi_email=config.pi_email;
          nc_root.do_name=config.do_name;
          nc_root.do_email=config.do_email;
          nc_root.ds_name=config.ds_name;
          nc_root.ds_email=config.ds_email;

          nc_root.instrument_number=config.instr_number;
          nc_root.instrument_channel=config.instr_channel;
          nc_root.instrument_type=config.instr_type;
          nc_root.station_name=config.station_name;
          nc_root.campaign_name=config.campaign_name;
          nc_root.measurement_funding_source='';
          nc_root.project_name=config.project;
          nc_root.file_name_prefix=config.prefix;
          nc_root.file_type="L1";
          nc_root.l1_file_version="fv"+file_version;

          fileName=os.path.splitext(os.path.basename(ncFile))[0];
          fileFields=fileName.split('-');
          nc_root.time_coverage_start=fileFields[7];
          nc_root.time_coverage_end=fileFields[8];

          # Create dimensions

          nc_root.createDimension('number_of_records',nrecords);
          nc_root.createDimension('detector_size',detsize);
          nc_root.createDimension('tp_level_size',tp.nlevels);
          nc_root.createDimension('tp_time_size',tp.ntime);
          nc_root.createDimension('dim1_size',1);
          nc_root.createDimension('datetime_size',7);
          nc_root.createDimension('cloud_size',cloud.cloud_size);
          nc_root.createDimension('aerosol_time_size',aod.ntime);
          nc_root.createDimension('aerosol_wavelength_size',aod.nwve);

          if (slfSize[0]>0):

             nc_slf_x=nc_root.createDimension('slit_dimx',slfSize[0]);
             nc_slf_y=nc_root.createDimension('slit_dimy',slfSize[1] if len(slfSize)>=2 else 1);

          # Create groups
          # TODO : create groups and variables only if there are data ???

          nc_instrument=nc_root.createGroup('INSTRUMENT_LOCATION');
          nc_ancillary=nc_root.createGroup('ANCILLARY');
          nc_key=nc_root.createGroup('KEYDATA');
          nc_radiance=nc_root.createGroup('RADIANCE');

          # Create instrument variables

          netCDF_NewVariable(nc_instrument,'latitude',float32,('dim1_size',),'degree_north','latitude of the instrument (positive north)',fillValue,value=float(config.instr_latitude),standard_name='latitude');
          netCDF_NewVariable(nc_instrument,'longitude',float32,('dim1_size',),'degree_east','longitude of the instrument (positive east)',fillValue,value=float(config.instr_longitude),standard_name='longitude');
          netCDF_NewVariable(nc_instrument,'altitude',float32,('dim1_size',),'m','altitude of the instrument above sea level',fillValue,float(config.instr_altitude_asl),standard_name='altitude');
          netCDF_NewVariable(nc_instrument,'altitude_of_station',float32,('dim1_size',),'m','altitude of the station above sea level',fillValue,float(config.station_altitude_asl),standard_name='ground_level_altitude');

          # Create ancillary variables

          nc_meteo=nc_ancillary.createGroup('METEOROLOGICAL_DATA');
          nc_surface=nc_ancillary.createGroup('SURFACE_DATA');
          nc_aerosol=nc_ancillary.createGroup('AEROSOL_DATA');

          nc_tp_pressure=nc_meteo.createGroup('TEMPERATURE_PRESSURE');

          netCDF_NewVariable(nc_tp_pressure,'meteo_time',float32,('dim1_size','tp_time_size'),'hour','UT time in fractional hour',fillValue,tp.meteo_time);
          netCDF_NewVariable(nc_tp_pressure,'altitude_level',float32,('tp_level_size',),'km','altitudes levels above sea level',fillValue,value=tp.altitude_level);
          netCDF_NewVariable(nc_tp_pressure,'pressure',float32,('tp_level_size','tp_time_size'),'hPa','pressure profile',fillValue,value=tp.pressure);
          netCDF_NewVariable(nc_tp_pressure,'temperature',float32,('tp_level_size','tp_time_size'),'K','temperature profile',fillValue,value=tp.temperature);

          nc_cloud=nc_meteo.createGroup('CLOUD_INFORMATION');

          netCDF_NewVariable(nc_cloud,'cloud_time',float32,('cloud_size'),'hour','UT time in fractional hour',fillValue,value=cloud.cloud_time);
          netCDF_NewVariable(nc_cloud,'cloud_coverage',float32,('cloud_size'),'percent','cloud coverage',fillValue,value=cloud.cloud_coverage,standard_name='cloud_area_fraction');
          netCDF_NewVariable(nc_cloud,'cloud_height',float32,('cloud_size'),'km','cloud height',fillValue,value=cloud.cloud_height,standard_name='cloud_base_altitude');
          if not isnan(surfaceAlbedo):  # MMF July 2018
            netCDF_NewVariable(nc_surface,'surface_albedo',float32,('dim1_size'),'1','surface albedo',fillValue,value=surfaceAlbedo,standard_name='surface_albedo');
          netCDF_NewVariable(nc_aerosol,'aerosol_time',float32,('aerosol_time_size',),'hour','UT time in fractional hour',fillValue,value=aod.time);
          netCDF_NewVariable(nc_aerosol,'aerosol_wavelength',float32,('dim1_size','aerosol_wavelength_size'),'nm','wavelength (in air)',fillValue,value=aod.wavelength);
          netCDF_NewVariable(nc_aerosol,'aerosol_optical_depth',float32,('aerosol_time_size','aerosol_wavelength_size'),'1','aerosol optical depth',fillValue,value=aod.aod,standard_name='atmosphere_optical_thickness_due_to_ambient_aerosol_particles');
          netCDF_NewVariable(nc_aerosol,'angstrom_exponent',float32,('aerosol_time_size'),'1','Angstrom exponent',fillValue,value=aod.alpha,standard_name='angstrom_exponent_of_ambient_aerosol_in_air');
          netCDF_NewVariable(nc_aerosol,'asymmetry_factor',float32,('aerosol_time_size','aerosol_wavelength_size'),'1','asymmetry factor',fillValue,value=aod.asymmetry);
          netCDF_NewVariable(nc_aerosol,'single_scattering_albedo',float32,('aerosol_time_size','aerosol_wavelength_size'),'1','single scattering albedo',fillValue,value=aod.albedo,standard_name='single_scattering_albedo_in_air_due_to_ambient_aerosol_particles');

          # Create key data

          # TODO : add slit function and reference file names as attributes ?

          if slfSize[0]>0 :

             nc_slit=nc_key.createGroup('SLIT_FUNCTION');

             netCDF_NewVariable(nc_slit,'slit_function_relative_wavelength',float32,('slit_dimx',),'nm','relative wavelength grid',fillValue,value=slfX);                                              # relative wavelengths
             netCDF_NewVariable(nc_slit,'slit_function_measured_wavelength',float32,('dim1_size','slit_dimy'),'nm','wavelengths at which the slit function has been measured',fillValue,value=slfY);   # measured wavelengths
             netCDF_NewVariable(nc_slit,'slit_function',float32,('slit_dimx','slit_dimy'),'1','measured slit function(s)',fillValue,value=slf);                                                          # measured slit functions

          if refSize[0]==detsize and refSize[1]>=2:

             nc_ref=nc_key.createGroup('REFERENCE_SPECTRUM');

             netCDF_NewVariable(nc_ref,'reference_wavelength',float32,('detector_size',),'nm','reference grid (in air)',fillValue,value=ref[:,0]);
             netCDF_NewVariable(nc_ref,'reference_spectrum',float32,('detector_size',),'1','reference spectrum',fillValue,value=ref[:,1]);

          # Create geodata variables

          nc_data=nc_radiance.createGroup('OBSERVATIONS');
          nc_geodata=nc_radiance.createGroup('GEODATA');

          #   Azimuth should be in [0..360] range

          L1_data.radiance.geodata.viewingAzimuth=(L1_data.radiance.geodata.viewingAzimuth+viewingAzimuthCorrection)%360.;
          L1_data.radiance.geodata.solarAzimuthAngle=(L1_data.radiance.geodata.solarAzimuthAngle+solarAzimuthCorrection)%360.;

          # --- mandatory fields

          netCDF_NewVariable(nc_geodata,'viewing_elevation_angle',float32,('number_of_records',),'degree','viewing elevation angle',fillValue,value=L1_data.radiance.geodata.viewingElevation);
          netCDF_NewVariable(nc_geodata,'viewing_azimuth_angle',float32,('number_of_records',),'degree','viewing azimuth angle 0..360, measured towards the east, from north',fillValue,value=L1_data.radiance.geodata.viewingAzimuth,standard_name='sensor_azimuth_angle');
          netCDF_NewVariable(nc_geodata,'solar_zenith_angle',float32,('number_of_records',),'degree','solar zenith angle',fillValue,value=L1_data.radiance.geodata.solarZenithAngle,standard_name='solar_zenith_angle');
          netCDF_NewVariable(nc_geodata,'solar_azimuth_angle',float32,('number_of_records',),'degree','solar azimuth angle, 0..360, measured towards the east, from north',fillValue,value=L1_data.radiance.geodata.solarAzimuthAngle,standard_name='solar_azimuth_angle');

          # netCDF_NewVariable(nc_geodata,'moon_zenith_angle',float32,('number_of_records',),'degree','moon zenith angle',fillValue,value=L1_data.radiance.geodata.solarZenithAngle);
          # netCDF_NewVariable(nc_geodata,'moon_azimuth_angle',float32,('number_of_records',),'degree','moon azimuth angle, 0..360, measured towards the east, from north',fillValue,value=L1_data.radiance.geodata.solarAzimuthAngle);

          netCDF_NewVariable(nc_data,'exposure_time',float32,('number_of_records',),'s','exposure time',fillValue,value=L1_data.radiance.observations.exposureTime);
          netCDF_NewVariable(nc_data,'number_of_coadded_spectra',int16,('number_of_records',),'1','number of co-added spectra',-1,value=L1_data.radiance.observations.nacc);
          netCDF_NewVariable(nc_data,'measurement_type',int16,('number_of_records',),'1','measurement type : 1-offaxis, 2-direct sun, 3-zenith, 7-almucantar, 11-horizon, 12-direct moon',-1,value=L1_data.radiance.observations.measureType);

                             # not necessary anymore longDate=netCDF_NewVariable(nc_meta,'long_datetime',str,('number_of_records',),'1','UT date and time in YYYYMMDDhhmmss.ms format','000000000000.00');
                             # fracDay=netCDF_NewVariable(nc_meta,'fractional_day',float32,('number_of_records',),'UT calendar day','fractional calendar day');
                             # fracTime=netCDF_NewVariable(nc_meta,'fractional_time',float32,('number_of_records',),'UT fractional hours','fractional time');
                             # fracStartTime=netCDF_NewVariable(nc_meta,'fractional_start_time',float32,('number_of_records',),'UT fractional hours','fractional start time');
                             # fracEndTime=netCDF_NewVariable(nc_meta,'fractional_end_time',float32,('number_of_records',),'UT fractional hours','fractional end time');

          # --- optional fields (for example, for mobile measurements)

                             # --- optional latitude=netCDF_NewVariable(nc_meta,'latitude',float32,('number_of_records',),'degrees_north','latitude (positive north)');
                             # --- optional longitude=netCDF_NewVariable(nc_meta,'longitude',float32,('number_of_records',),'degrees_east','longitude (positive east)');

          netCDF_NewVariable(nc_data,'total_acquisition_time',float32,('number_of_records',),'s','total acquisition time (the total time the detector is exposed to light to produce the spectrum)',fillValue,value=L1_data.radiance.observations.totalAcqTime);
          netCDF_NewVariable(nc_data,'total_measurement_time',float32,('number_of_records',),'s','total measurement time (should be the time difference in seconds between UTC time start and UTC time stop)',fillValue,value=L1_data.radiance.observations.totalMeasTime);

          # date and time

          for i in range(0,nrecords):

             dateCurrent=L1_data.radiance.observations.utTime[i,:];
             dateStart=L1_data.radiance.observations.utStart[i,:];
             dateEnd=L1_data.radiance.observations.utEnd[i,:];

             d1=datetime(dateStart[0],dateStart[1],dateStart[2],dateStart[3],dateStart[4],dateStart[5]);
             d=datetime(dateCurrent[0],dateCurrent[1],dateCurrent[2],dateCurrent[3],dateCurrent[4],dateCurrent[5]);
             d2=datetime(dateEnd[0],dateEnd[1],dateEnd[2],dateEnd[3],dateEnd[4],dateEnd[5]);

             if d1>d:
                d1=d1-timedelta(days=1);
                dateStart[0]=d1.year;
                dateStart[1]=d1.month;
                dateStart[2]=d1.day;

             if d2<d:
                d2=d2+timedelta(days=1);
                dateEnd[0]=d1.year;
                dateEnd[1]=d1.month;
                dateEnd[2]=d1.day;

             L1_data.radiance.observations.utTime[i,:]=dateCurrent;
             L1_data.radiance.observations.utStart[i,:]=dateStart;
             L1_data.radiance.observations.utEnd[i,:]=dateEnd;

          netCDF_NewVariable(nc_data,'datetime',int16,('number_of_records','datetime_size'),'1','measurement date and time (UT YYYY,MM,DD,hh,mm,ss,ms)',-1,value=L1_data.radiance.observations.utTime);
          netCDF_NewVariable(nc_data,'datetime_start',int16,('number_of_records','datetime_size'),'1','start date and time (UT YYYY,MM,DD,hh,mm,ss,ms)',-1,value=L1_data.radiance.observations.utStart);
          netCDF_NewVariable(nc_data,'datetime_end',int16,('number_of_records','datetime_size'),'1','end date and time (UT YYYY,MM,DD,hh,mm,ss,ms)',-1,value=L1_data.radiance.observations.utEnd);

          # Create measurements variables

          netCDF_NewVariable(nc_data,'wavelength',float32,('number_of_records','detector_size'),'nm','wavelength grid (in air)',fillValue,value=L1_data.radiance.observations.wavelength);
          netCDF_NewVariable(nc_data,'radiance',float32,('number_of_records','detector_size'),'1','sum of co-added spectra (count number)',fillValue,value=L1_data.radiance.observations.spectra);
          netCDF_NewVariable(nc_data,'radiance_error',float32,('number_of_records','detector_size'),'1','error on sum of co-added spectra (count number)',fillValue);
          netCDF_NewVariable(nc_data,'radiance_quality_flag',int16,('number_of_records','detector_size'),'1','pixel quality flags (currently 0 or 1 but probably to complete later)',1);

                             # longDate[:]=asarray(datestr,dtype='U');

          # Close file

          nc_root.close();

    return ncFile;

def netCDF_L1_ExtractSlitFunction(ncFile,slfFile):
    """
    Extract the slit function from a radiance L1 file and save it in ASCII for QDOAS

    Parameters
    ----------

    ncFile     : the name of the input file
    slfFile    : name of the output slit function file (ASCII format)                          \n

    Returns
    -------

    0 if success, 1 on error
    """

    functionName="netCDF_L1_ExtractSlitFunction"
    rc=1;

    if not os.path.isfile(ncFile):
       print('['+functionName+'] '+ncFile+' not found !!!');
    elif not os.path.isdir(os.path.dirname(slfFile)):
       print('['+functionName+'] '+os.path.dirname(slfFile)+' not a directory !!!');
    else:

      try:

        nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');

        if 'KEYDATA' not in nc.groups :
           print('['+functionName+'] '+ncFile+' do not contain KEYDATA group');
        elif 'SLIT_FUNCTION' not in nc.groups['KEYDATA'].groups:
           print('['+functionName+'] '+ncFile+' do not contain SLIT_FUNCTION group in KEYDATA group');
        elif 'SLIT_FUNCTION' not in nc.groups['KEYDATA'].groups or \
             'slit_function_relative_wavelength' not in nc.groups['KEYDATA'].groups['SLIT_FUNCTION'].variables or \
             'slit_function_measured_wavelength' not in nc.groups['KEYDATA'].groups['SLIT_FUNCTION'].variables :

           print('['+functionName+'] '+ncFile+' : some variables are missing in the SLIT_FUNCTION group');

        else:

           slit=nc.groups['KEYDATA'].groups['SLIT_FUNCTION'].variables['slit_function']
           slitw=nc.groups['KEYDATA'].groups['SLIT_FUNCTION'].variables['slit_function_relative_wavelength']
           slitc=nc.groups['KEYDATA'].groups['SLIT_FUNCTION'].variables['slit_function_measured_wavelength'];

           slitf=c_[asarray(slitw),asarray(slit)] if (slitc.size==1) else r_[c_[0.,asarray(slitc)],c_[asarray(slitw),asarray(slit)]]
           savetxt(slfFile,slitf,fmt='%16.8e ',newline='\n',delimiter=' ');

           rc=0;

      except OSError:
        print('['+functionName+'] Can not open '+ncFile+' in read mode');

    return 0;

def netCDF_L1_ExtractReference(ncFile,refFile):
    """
    Extract the reference from a radiance L1 file and save it in ASCII for QDOAS

    If the reference is in the keydata, it should extract this reference
    otherwise, the zenith spectrum at noon will be extracted

    Parameters
    ----------

    ncFile     : the name of the input file
    refFile    : name of the reference spectrum file (ASCII format)                          \n

    Returns
    -------

    0 if success, 1 on error
    """

    functionName="netCDF_L1_ExtractReference"
    rc=1;

    if not os.path.isfile(ncFile):
       print('['+functionName+'] '+ncFile+' not found !!!');
    elif not os.path.isdir(os.path.dirname(refFile)):
       print('['+functionName+'] '+os.path.dirname(refFile)+' not a directory !!!');
    else:

      try:

        nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');

        if 'number_of_records' not in nc.dimensions or 'detector_size' not in nc.dimensions:
           print('['+functionName+'] '+ncFile+' : some dimensions are missing (number_of_records or detector_size)');
        elif (nc.dimensions['number_of_records'].size==0) or (nc.dimensions['detector_size'].size==0):
           print('['+functionName+'] '+ncFile+' : wrong dimensions (number_of_records or detector_size)');
        elif 'KEYDATA' in nc.groups and 'REFERENCE_SPECTRUM' in nc.groups['KEYDATA'].groups :

          if 'RADIANCE' not in nc.groups and ('GEODATA' not in nc.groups['RADIANCE'].groups or 'OBSERVATIONS' not in nc.groups['RADIANCE'].groups) :
             print('['+functionName+'] '+ncFile+' : some groups are missing (KEYDATA or OBSERVATIONS)');

          elif 'measurement_type' not in nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables or \
               'solar_zenith_angle' not in nc.groups['RADIANCE'].groups['GEODATA'].variables or \
               'radiance' not in nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables or \
               'wavelength' not in nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables:

             print('['+functionName+'] '+ncFile+' : some variables are missing in the OBSERVATIONS or GEODATA group (part of RADIANCE group)');
          else:
             meas_type=asarray(nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables['measurement_type']);
             zen=asarray(nc.groups['RADIANCE'].groups['GEODATA'].variables['solar_zenith_angle']);

             i=[i for (i,mt) in enumerate(meas_type) if mt==QDOAS_MeasType_zenith];

             if len(i):

                zeni=zen[i];
                j=[j for (j,z) in enumerate(zeni) if z==min(zeni)];

                radiances=asarray(nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables['radiance']);
                wavelengths=asarray(nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables['wavelength']);

                wve=wavelengths[i[j[0]],:];
                spe=zeros(nc.dimensions['detector_size'].size);

                for jj in j:
                   spe=spe+radiances[i[jj]];

                spe=spe/len(j);
                ref=c_[wve,spe];

                savetxt(refFile,ref,fmt='%16.8e %24.16e',newline='\n',delimiter=' ');

                rc=0;

        else:
           print(nc.groups['KEYDATA'])
           print('['+functionName+'] Extraction of reference from the key data not implemented yet');

      except OSError:
        print('['+functionName+'] Can not open '+ncFile+' in read mode');

    return rc;

def netCDF_L1_GetDetectorSize(ncFile):
    """
    Return the size of the detector

    Parameters
    ----------

    ncFile     : the name of the input file

    Returns
    -------

    the size of the detector
    """

    functionName="netCDF_L1_GetDetectorSize"
    detectorSize=0;

    if not os.path.isfile(ncFile):
       print('['+functionName+'] '+ncFile+' not found !!!');
    else:

      try:

        nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');

        if 'detector_size' not in nc.dimensions:
           print('['+functionName+'] '+ncFile+' : dimensions detector_size not found');
        else:
           detectorSize=nc.dimensions['detector_size'].size;

      except OSError:
        print('['+functionName+'] Can not open '+ncFile+' in read mode');

    return detectorSize;

def netCDF_L1_GetCalibRange(ncFile,delta=3.):
    """
    Return the wavelength calibration limits with a security delta

    Parameters
    ----------

    ncFile     : the name of the input file
    delta      : the security delta to avoid detector edge effects.

    Returns
    -------

    (lambdaMin,lambdaMax)
    """

    functionName="netCDF_L1_GetCalibRange"
    lambdaMin=0.;
    lambdaMax=0.;

    if not os.path.isfile(ncFile):
       print('['+functionName+'] '+ncFile+' not found !!!');
    else:

      try:

        nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');

        if 'detector_size' not in nc.dimensions:
           print('['+functionName+'] '+ncFile+' : missing dimension (detector_size)');
        elif (nc.dimensions['detector_size'].size==0):
           print('['+functionName+'] '+ncFile+' : wrong dimensions (detector_size)');
        elif 'RADIANCE' not in nc.groups :
           print('['+functionName+'] '+ncFile+' : missing group (RADIANCE)');
        elif 'OBSERVATIONS' not in nc.groups['RADIANCE'].groups :
           print('['+functionName+'] '+ncFile+' : missing group (RADIANCE.OBSERVATIONS)');
        elif 'wavelength' not in nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables:
           print('['+functionName+'] '+ncFile+' : missing variables in the OBSERVATION group (wavelength)');
        else:
           detectorSize=nc.dimensions['detector_size'].size;

           wavelengths=asarray(nc.groups['RADIANCE'].groups['OBSERVATIONS'].variables['wavelength']);
           lambdaMin=ceil(wavelengths[0][0]+delta);
           lambdaMax=floor(wavelengths[0][detectorSize-1]-delta);

      except OSError:
        print('['+functionName+'] Can not open '+ncFile+' in read mode');

    return (lambdaMin,lambdaMax);

def netCDF_L1_SetFlux(lambdaMin,lambdaMax,possibleFlux):
    """
    Return the fluxes possible within a specified wavelength calibration range

    Parameters
    ----------

    lambdaMin,lambdaMax : the limits of the wavelength calibration range
    possibleFlux        : a list of fluxes to test if there are within the wavelength range.

    Returns
    -------

    newFlux : a subset of possibleFlux
    """

    functionName="netCDF_L1_SetFlux"

    newFlux=[];

    for flux in possibleFlux:
       if flux>lambdaMin and flux<lambdaMax:
          newFlux.append(flux);

    return newFlux;

# Example of call
#
# netCDF_L1_Convert('/bira-iasb/projects/FRM4DOAS/tmp/demo/config/level1/BXVIS_1670_1_ccdeev2netcdf.ini',
#                   '/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Brussels/MAXDOAS/Data/spectra/VIS/BXVIS/DSC/2018/BXVIS_20180526_146.DSC',
#                   '/bira-iasb/projects/FRM4DOAS/tmp/format/new_format',
#                   '/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Brussels/MAXDOAS/Data/spectra/VIS/BXVIS/Calib/BXVIS_20170130_030.REF',
#           slfFile='/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Campaigns/CINDI_2016/Data/spectra/VIS/Calib/slit_435_CBVIS.slf',
#           clbFile='/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Brussels/MAXDOAS/Data/spectra/VIS/BXVIS/Calib/BXVIS_20170128_028_FromKurucz.clb',
#                    file_version='2',slfWavelength=435.,file_format='CCDEEV',commentChar=';',
#         cloudFile='/bira-iasb/projects/FRM4DOAS/tmp/format/new_format/cloud.txt',
#            tpFile='/bira-iasb/projects/FRM4DOAS/tmp/format/new_format/atm_std6.dat',
#           aodFile='/bira-iasb/projects/FRM4DOAS/tmp/format/new_format/20180526_20180526_Brussels.lev15');

# netCDF_L1_Convert('/bira-iasb/projects/FRM4DOAS/tmp/demo/config/level1/cindi2_004_1_asc2netcdf.ini',
#                   '/bira-iasb/projects/FRM4DOAS/data/spectra/campaigns/cindi2/bira_4/bira_maxdoas_4_uv_spectra_cindi2_20160924_v1.asc',
#                   '/bira-iasb/projects/FRM4DOAS/tmp/format/new_format',
#                   '/bira-iasb/projects/FRM4DOAS/data/results/campaigns/cindi2/tests/bira_4/bira_maxdoas_4_uv_spectra_cindi2_20160924_v1.ref',
#           slfFile='/bira-iasb/projects/FRM4DOAS/data/slitfunctions/campaigns/cindi2/bira_4/bira_maxdoas_4_uv_sltfct_cindi2_20160924_v1.asc',
#           clbFile='/bira-iasb/projects/FRM4DOAS/data/results/campaigns/cindi2/tests/bira_4/bira_maxdoas_4_uv_spectra_cindi2_20160924_v1.ref',
#                    file_version='2',file_format='ASCII',commentChar='#',
#         cloudFile='/bira-iasb/projects/FRM4DOAS/tmp/format/new_format/cloud.txt',
#            tpFile='/bira-iasb/projects/FRM4DOAS/tmp/format/new_format/atm_std6.dat',
#           aodFile='/bira-iasb/projects/FRM4DOAS/tmp/format/new_format/20180526_20180526_Brussels.lev15',
#     slfWavelength=346.);

