#! /usr/bin/env python

import os

from tools.doas_lib import *
from numpy import *
from pylab import *
from datetime import *
from netCDF4 import Dataset

import struct
import scipy.io as sio

# import scipy.linalg

import matplotlib

matplotlib.rc('xtick', labelsize=12)
matplotlib.rc('ytick', labelsize=12)



class CCD_spectrum :

    """
    Definition of a spectrum
    ------------------------

         recId             : record number if known (-1 if unknown)                                                                     \n
         header            : general information on the record (see complete description below)                                         \n
         spectrum          : spectrum (average of tracks) or matrix of spectra if tracks are saved individually                         \n
         spectrumCorrected : if offsets are measured and saved in the spectra file, spectra are corrected by this offset                \n

         For measurements with the integration time automatically calculated (header.nTint <> 0) :                                      \n

               ccdTabTint  : array with the different integration times used for the current acquisition                                \n
               ccdTabNTint : for each integration time in the previous table, the number of scans used for this integration time        \n
               ccdSpecMax  : the variation of the signal vs the number of scans                                                         \n

    Description of the header in Matlab
    -----------------------------------

         ==  ====================  ===========  ======  =====================================================================
         no  name                  type         length  comment
         ==  ====================  ===========  ======  =====================================================================
          1  'exposureTime'        'real*8'         1   exposure time
          2  'totalTime'           'int32'          1   the total measurement time
          3  'lockedTemperature'   'int32'          1   1 if temperature is locked, 0 otherwise
          4  'targetTemperature'   'real*8'         1   target temperature
          5  'currentTemperature'  'real*8'         1   current temperature
          6  'roiWveStart'         'int32'          1   region of interest
          7  'roiWveEnd'           'int32'          1
          8  'roiWveGroup'         'int32'          1
          9  'roiSlitStart'        'int32'          1
         10  'roiSlitEnd'          'int32'          1
         11  'roiSlitGroup'        'int32'          1
         12  'da_year'             'int16'          1   year part of the measurement date
         13  'da_day'              'int8'           1   day part of the measurement date
         14  'da_month'            'int8'           1   month part of the measurement date
         15  'ti_min'              'int8'           1   min part of the measurement time
         16  'ti_hour'             'int8'           1   hour part of the measurement time
         17  'ti_hund'             'int8'           1   hundredth of sec part of the measurement time
         18  'ti_sec'              'int8'           1   secunds part of the measurement time
         19  'Tm'                  'real*8'         1   measurement date and time in seconds
         20  'Zm'                  'real*8'         1   solar zenith angle
         21  'Azimuth'             'real*8'         1   solar azimuth angle
         22  'nrej'                'int32'          1   number of rejected spectra
         23  'nacc'                'int32'          1   number of accumulation
         24  'saveTracks'          'int32'          1   1 if all tracks are saved, 0 if only the mean is saved
         25  'vipCorrection'       'int32'          1   1 if a interpixel variability correction has been applied 0 otherwise
         26  'nlCorrection'        'int32'          1   1 if a nonlinearity correction has been applied 0 otherwise
         27  'nTint'               'int16'          1   number of integration times
         28  'brusagElevation'     'real*8'         1   elevation angle for off axis measurements
         29  'filterUnused'        'int32'          1   the name of the filter used if any
         30  'filterNumber'        'int32'          1   filter number
         31  'headTemperature'     'real*8'         1   head temperature
         32  'alsFlag'             'int16'          1   als measurements
         33  'scanIndex'           'int16'          1   als scan index
         34  'scanningAngle'       'real*8'         1   als scan angle
         35  'doubleFlag'          'int8'           1   1 if spectra are saved in double
         36  'brusagVersion'       'int8'           1   version of intra
         37  'diodes'              'real*4'         4   sun diodes
         38  'targetElevation'     'real*4'         1   requested elevation
         39  'targetAzimuth'       'real*4'         1   requested azimuth
         40  'saturatedFlag'       'int8'           1   saturated pixel
         41  'ignoredFlag'         'int8'           1   ignored byte
         42  'wve1'                'real*4'         1   wavelength 1 for flux 1
         43  'wve2'                'real*4'         1   wavelength 2 for flux 2
         44  'flux1'               'real*4'         1   flux 1 calculated on wavelength 1
         45  'flux2'               'real*4'         1   flux 2 calculated on wavelength 2
         46  'ignored'             'int8'         906   ignored bytes
         47  'jewell'              'real*4'         1   inclinometer
         48  'compass'             'real*4'         1   for airborne measurements
         49  'pitch'               'real*4'         1   for airborne measurements
         50  'roll'                'real*4'         1   for airborne measurements
         51  'start_ti_min'        'int8'           1   min part of the starting time
         52  'start_ti_hour'       'int8'           1   hour part of the starting time
         53  'start_ti_hund'       'int8'           1   hundredth of sec
         54  'start_ti_sec'        'int8'           1   secunds part
         55  'end_ti_min'          'int8'           1   min part of the ending time
         56  'end_ti_hour'         'int8'           1   hour part of the ending time
         57  'end_ti_hund'         'int8'           1   hundredth of sec part of the ending time
         58  'end_ti_sec'          'int8'           1   secunds part of the ending time
         59  'brusagAzimuth'       'real*8'         1   brusag azimuth
         60  'measureType'         'int32'          1   type of measurement
         ==  ====================  ===========  ======  =====================================================================

    Equivalent of the header in Python
    ----------------------------------

         '=d2i2d6ih6B3d5ihd2id2hd2B6f2B4f906s4f8Bdi'

    with :

         =========  ======================  ========  ==========
         Character  Byte order              Size      Alignment
         =========  ======================  ========  ==========
         @          native                  native    native
         =          native                  standard  none
         <          little-endian           standard  none
         >          big-endian              standard  none
         !          network (= big-endian)  standard  none
         =========  ======================  ========  ==========

         ======  ==============  ==================  ======================
         Format  C Type          Python type         Standard size  (byte)}     \n
         ======  ==============  ==================  ======================
         x       pad             byte                no value                   \n
         c       char            string of length 1  1                          \n
         b       signed char     integer             1                          \n
         B       unsigned char   integer             1                          \n
         ?       _Bool           bool                1                          \n
         h       short           integer             2                          \n
         H       unsigned short  integer             2                          \n
         i       int             integer             4                          \n
         I       unsigned int    integer             4                          \n
         l       long            integer             4                          \n
         L       unsigned long   integer             4                          \n
         q       long long       integer             8                          \n
         Q       unsigned long   long  integer       8                          \n
         f       float           float               4                          \n
         d       double          float               8                          \n
         s       char[]          string                                         \n
         p       char[]          string                                         \n
         ======  ==============  ==================  ======================
    """

    def __init__(self,*args):

       self.recId=args[0]                                                       # record number

       self.header=args[1]                                                      # header
       self.spectrum=args[2]                                                    # original spectrum
       self.spectrumCorrected=args[3]                                           # spectrum corrected by the offset if the offset is present in the file

       if self.header['nTint']:

          self.tint=args[4];                                                    # vector with the used integration times
          self.ntint=args[5];                                                   # per integration time in the previous vector, the number of times this integration has been used
          self.specMax=args[6];                                                 # variation of the signal max along the scans

CCD_fields= ('exposureTime',      \
             'totalTime',         \
             'lockedTemperature', \
             'targetTemperature', \
             'currentTemperature',\
             'roiWveStart',       \
             'roiWveEnd',         \
             'roiWveGroup',       \
             'roiSlitStart',      \
             'roiSlitEnd',        \
             'roiSlitGroup',      \
             'da_year',           \
             'da_day',            \
             'da_month',          \
             'ti_min',            \
             'ti_hour',           \
             'ti_hund',           \
             'ti_sec',            \
             'Tm',                \
             'Zm',                \
             'Azimuth',           \
             'nrej',              \
             'nacc',              \
             'saveTracks',        \
             'vipCorrection',     \
             'nlCorrection',      \
             'nTint',             \
             'brusagElevation',   \
             'filterUnused',      \
             'filterNumber',      \
             'headTemperature',   \
             'alsFlag',           \
             'scanIndex',         \
             'scanningAngle',     \
             'doubleFlag',        \
             'brusagVersion',     \
             'diode_1',           \
             'diode_2',           \
             'diode_3',           \
             'diode_4',           \
             'targetElevation',   \
             'targetAzimuth',     \
             'saturatedFlag',     \
             'ignoredFlag',       \
	            'wve1',              \
	            'wve2',              \
	            'flux1',             \
	            'flux2',             \
             'ignored',           \
             'jewell',            \
             'compass',           \
             'pitch',             \
             'roll',              \
             'start_ti_min',      \
             'start_ti_hour',     \
             'start_ti_hund',     \
             'start_ti_sec',      \
             'end_ti_min',        \
             'end_ti_hour',       \
             'end_ti_hund',       \
             'end_ti_sec',        \
             'brusagAzimuth',     \
             'measureType');

CCD_headerFmt = '=d2i2d6ih6B3d5ihd2id2hd2B6f2B4f906s4f8Bdi' # '=d2i2d6ih6B3d5ihd2id2hd2B6f2B926s3f8Bdi'
CCD_headerLen = struct.calcsize(CCD_headerFmt)
CCD_headerUnpack = struct.Struct(CCD_headerFmt).unpack_from

class CCD_info(object):
    def __init__(self,ut,measType,tint,totalTime,sza,sazim,nrej,nscans,ntint,trackerElevation,trackerAzimuth,filter,targetElev,targetAzim,startTime,endTime,jewell,flux,saturatedFlag):

       self.ut=ut;
       self.measType=measType;
       self.tint=tint;
       self.totalTime=totalTime;
       self.sza=sza;
       self.sazim=sazim;
       self.nrej=nrej;
       self.nscans=nscans;
       self.ntint=ntint;
       self.trackerElevation=trackerElevation;
       self.trackerAzimuth=trackerAzimuth;
       self.filter=filter;
       self.targetElevation=targetElev;
       self.targetAzimuth=targetAzim;
       self.startTime=startTime;
       self.endTime=endTime;
       self.jewell=jewell;
       self.flux=flux;
       self.saturatedFlag=saturatedFlag;

# CCD measurement types

CCD_MT_offaxis=QDOAS_MeasType_offaxis;
CCD_MT_ds=QDOAS_MeasType_ds;
CCD_MT_zenith=QDOAS_MeasType_zenith;
CCD_MT_drk=QDOAS_MeasType_dark;
CCD_MT_almucantar=QDOAS_MeasType_almucantar;
CCD_MT_offset=QDOAS_MeasType_offset;
CCD_MT_azimuth=QDOAS_MeasType_azimuth;
CCD_MT_principalPlane=QDOAS_MeasType_principalplane;
CCD_MT_horizon=QDOAS_MeasType_horizon;

# Control variables

CCD_Debug=0

def CCD_ReadNextRecord(f,recordId):

    """
    CCD_record=CCD_ReadNextRecord(f,recordId)

    return the next CCD record whatever the measurement type (observation, offset, dark current

    Parameters
    ----------

    f : pointer to the file to read

    Returns
    -------

    CCD_record : the next record

    See also
    --------

    CCD_spectrum,CCD_ReadSpe
    """

    # Read header

    bytesPacked = f.read(CCD_headerLen)                                         # read CCD_headerLen bytes
    if not bytesPacked: return []
    bytesUnpacked = CCD_headerUnpack(bytesPacked)                               # decomposition of the bytes pack according to the format string

    # Associate the list of header fields with the list of items in s

    header=dict(  zip(CCD_fields, bytesUnpacked) )                              # map the two lists into a dictionary

    # Calculate the size of the data to load

    ccdX=(header['roiWveEnd']-header['roiWveStart']+1)/header['roiWveGroup'];
    ccdY=(header['roiSlitEnd']-header['roiSlitStart']+1)/header['roiSlitGroup'];

    if CCD_Debug:
       print(sprintf('%02d:%02d:%02d, %.3f, %.2f, %d, %d, %.2f, %.2f',header['ti_hour'],header['ti_min'],header['ti_sec'],header['Zm'],header['exposureTime'],header['nacc'],header['measureType'],header['targetElevation'],header['brusagElevation']))

    spectraSize=ccdX*ccdY if header['saveTracks']==1 else ccdX;
    nTint=header['nTint']

    # Read spectrum

    spectrum=array(reshape(fromfile(f,dtype=uint16,count=int(spectraSize)),(int(ccdX),-1),order='F'),dtype=double);

    # For observations only (not for the dark currents neither offset)

    if nTint:

       ccdTabTint=array(fromfile(f,dtype=double,count=nTint),dtype=double);
       ccdTabNTint=array(fromfile(f,dtype=uint16,count=nTint),dtype=double);
       ccdSpecMax=array(fromfile(f,dtype=double,count=header['nacc']+header['nrej']),dtype=double);

       CCD_record=CCD_spectrum(recordId,header,spectrum,spectrum,ccdTabTint,ccdTabNTint,ccdSpecMax);

       if (header['saturatedFlag']==0):

         saturated=0;
         for i in arange(0,nTint):
            if ccdSpecMax[i]<1.e-6 or ccdSpecMax[i]>60000.:
               saturated=saturated+1;

         if (saturated==nTint):
            header['saturatedFlag']=1;

    else:

       CCD_record=CCD_spectrum(recordId,header,spectrum,spectrum);

    return CCD_record;

def CCD_SaveSpe(g,spe):

    """
    CCD_SaveSpe(spe)

    Save a spectrum record in the output file

    Parameters
    ----------

    g   : pointer to the output file
    spe : spectrum to save (CCD_spectrum structure)

    See also
    --------

    CCD_spectrum
    """

    header=spe.header;
    ccdX=header['roiWveEnd']-header['roiWveStart']+1
    ccdY=(header['roiSlitEnd']-header['roiSlitStart']+1)/header['roiSlitGroup'];

    bytesPacked=struct.pack(CCD_headerFmt, \
                            header['exposureTime'],      \
                            header['totalTime'],         \
                            header['lockedTemperature'], \
                            header['targetTemperature'], \
                            header['currentTemperature'],\
                            header['roiWveStart'],       \
                            header['roiWveEnd'],         \
                            header['roiWveGroup'],       \
                            header['roiSlitStart'],      \
                            header['roiSlitEnd'],        \
                            header['roiSlitGroup'],      \
                            header['da_year'],           \
                            header['da_day'],            \
                            header['da_month'],          \
                            header['ti_min'],            \
                            header['ti_hour'],           \
                            header['ti_hund'],           \
                            header['ti_sec'],            \
                            header['Tm'],                \
                            header['Zm'],                \
                            header['Azimuth'],           \
                            header['nrej'],              \
                            header['nacc'],              \
                            header['saveTracks'],        \
                            header['vipCorrection'],     \
                            header['nlCorrection'],      \
                            header['nTint'],             \
                            header['brusagElevation'],   \
                            header['filterUnused'],      \
                            header['filterNumber'],      \
                            header['headTemperature'],   \
                            header['alsFlag'],           \
                            header['scanIndex'],         \
                            header['scanningAngle'],     \
                            header['doubleFlag'],        \
                            header['brusagVersion'],     \
                            header['diode_1'],           \
                            header['diode_2'],           \
                            header['diode_3'],           \
                            header['diode_4'],           \
                            header['targetElevation'],   \
                            header['targetAzimuth'],     \
                            header['saturatedFlag'],     \
                            header['ignoredFlag'],       \
                            header['wve1'],              \
                            header['wve2'],              \
                            header['flux1'],             \
                            header['flux2'],             \
                            header['ignored'],           \
                            header['jewell'],            \
                            header['compass'],           \
                            header['pitch'],             \
                            header['roll'],              \
                            header['start_ti_min'],      \
                            header['start_ti_hour'],     \
                            header['start_ti_hund'],     \
                            header['start_ti_sec'],      \
                            header['end_ti_min'],        \
                            header['end_ti_hour'],       \
                            header['end_ti_hund'],       \
                            header['end_ti_sec'],        \
                            header['brusagAzimuth'],     \
                            header['measureType']);

    # s=struct.pack('uint16'*len(floats), *floats)
    #spectrum=array(reshape(spe.spectrumCorrected,(int(ccdX),-1),dtype=double);
    #ntint=reshape(array(spe.ntint+0.5,dtype=uint16),[header['nTint'],]);

    # !!! spectrum=array(spe.spectrumCorrected,dtype=uint16);

    spectrumSize=ccdX*ccdY if header['saveTracks']==1 else ccdX;
    spectrum=array(reshape(spe.spectrumCorrected,(int(spectrumSize),),order='F'),dtype=uint16);

    g.write(bytesPacked);

    # spectrum.tofile(g);
    g.write(spectrum);

    if header['nTint']:

       ntint=array(spe.ntint+0.5,dtype=uint16);

       g.write(spe.tint)
       g.write(ntint)
       g.write(spe.specMax);

def CCD_Plot(spe):

    """
    CCD_Plot(spe)

    Plot the spectrum

    Parameters
    ----------

    spe : spectrum to plot (CCD_spectrum structure)

    See also
    --------

    CCD_spectrum
    """

    header=spe.header;

    plot(spe.spectrumCorrected);

    if header['nTint']:
       str=sprintf('%d : %02d:%02d:%02d sza : %.2f elev : %.2f meas %d   %d x %.2f sec',spe.recId,header['ti_hour'],header['ti_min'],header['ti_sec'],header['Zm'],header['brusagElevation'],header['measureType'],header['nacc'],header['exposureTime']);
    else:
       str=sprintf('%d : %02d:%02d:%02d %d x %.2f sec',spe.recId,header['ti_hour'],header['ti_min'],header['ti_sec'],header['nacc'],header['exposureTime']);

    suptitle(str,fontsize=16,fontweight='bold');
    xlabel('Pixel number',fontsize=14,fontweight='bold')
    ylabel('Counts',fontsize=14,fontweight='bold')

    grid(True)

    while not waitforbuttonpress():
       pass

    clf()

def CCD_Average(spe):

    """
    CCD_Average(spe)

    Average all tracks

    Parameters
    ----------

    spe : spectrum to average (CCD_spectrum structure)

    Returns
    -------

    newSpe : the spectrum structure with spectra averaged on tracks

    See also
    --------

    CCD_spectrum
    """

    newSpe=spe;

    if spe.header['saveTracks']:                                                # all tracks are saved

       newSpe.spectrum=spe.spectrum.mean(1)
       newSpe.spectrumCorrected=spe.spectrumCorrected.mean(1)

       i=[j for (j,item) in enumerate(newSpe.spectrumCorrected) if item<=0.];

       if len(i)>0:
          newSpe.spectrumCorrected[i]=0.;

       newSpe.header['saveTracks']=0;
       newSpe.header['roiSlitEnd']=1;
       newSpe.header['roiSlitStart']=1;
       newSpe.header['roiSlitGroup']=1;

    return newSpe;

def CCD_ReadSpe(speFile,offsetFlag,averageFlag,plotFlag,measType,expected):

    """
    CCD_spe=CCD_ReadSpe(speFile,offsetFlag,plotFlag,measType)

    Read spectra measured by CCD detectors

    Parameters
    ----------

    speFile : the name of the spectra file to load                              \n
    offsetFlag : 1 to correct spectra by the offset if offset are measured      \n
                 2 to correct with the closest offset measurement               \n
    averageFlag : 1 to average spectra before returning                         \n
    plotFlag : 1 to plot the spectra                                            \n
    measType : use 0 for all spectra                                            \n
    expected : number of expected spectra (replicate the last one for missing spectra)

    Returns
    -------

    CCD_spe : the spectra retrieved from the file

    See also
    --------

    CCD_ReadNextRecord
    """

    CCD_spe=[]                                                                  # array of spectra to return
    CCD_speTmp=[]                                                               # array of spectra not corrected yet

    lastOffsetIdx=-1;
    newOffsetIdx=-1;
    offsetIdx=[];

    if not os.path.isfile(speFile):
        print('[CCD_ReadSpe] '+speFile+' not found !!!')
    else:

        f=open(speFile,"rb")

        nrec=0

        while True:

            # ----------------
            # READ NEXT RECORD
            # ----------------

            nrec=nrec+1
            offsetIdx.append(-1)

            spe=CCD_ReadNextRecord(f,nrec)
            if size(spe)==0: break;
            header=spe.header;

            if ((measType==0) or (offsetFlag and (header['measureType']==CCD_MT_offset)) or (header['measureType']==measType)):

               # -----------------
               # OFFSET CORRECTION
               # -----------------

               # Measurements other than zenith and off-axis measurements : correct with the previous offset

               if (offsetFlag==0) or (offsetFlag==2) or (header['measureType']!=CCD_MT_zenith and header['measureType']!=CCD_MT_offaxis and header['measureType']!=CCD_MT_drk):

                  if offsetFlag:

                     # Measurements other than zenith and off-axis measurements : correct with the previous offset

                     if (header['measureType']!=CCD_MT_offset) :
                        CCD_speTmp.append(spe);

                     else:

                        if newOffsetIdx!=-1 :
                           lastOffset=newOffset;

                        lastOffsetIdx=newOffsetIdx;
                        newOffset=spe.spectrum;
                        newOffsetIdx=nrec;

                        # Offset correction of records in the temporary list

                        for i,item in enumerate(CCD_speTmp):

                            if (item.header['nacc']>0):

                               if ((offsetFlag==1) or (lastOffsetIdx==-1) or (nrec-(lastOffsetIdx+1)-i<i+1) ) :
                                  item.spectrumCorrected=item.spectrum-newOffset;
                                  item.offsetIdx=newOffsetIdx;
                               else:
                                  item.spectrumCorrected=item.spectrum-lastOffset;
                                  item.offsetIdx=lastOffsetIdx;

                               offsetIdx[lastOffsetIdx+i+1]=item.offsetIdx;

                            else:
                               item.offsetIdx=-1;

                            if averageFlag:
                               CCD_speTmp[i]=CCD_Average(item);

                            CCD_spe.append(CCD_speTmp[i]);

                            if plotFlag:
                               CCD_Plot(CCD_speTmp[i]);

                        CCD_speTmp=[];


                  else:

                     if averageFlag:
                        spe=CCD_Average(spe);

                     CCD_spe.append(spe);

                     if plotFlag:
                        CCD_Plot(spe);

               # Zenith-sky and off-axis measurements : correct with the next offset measurement

               else:

                  CCD_speTmp.append(spe);


        f.close()

        if size(CCD_speTmp)>0:

           for i,item in enumerate(CCD_speTmp):

               if 'newOffset' in locals():

                  item.spectrumCorrected=item.spectrum-newOffset
                  item.offsetIdx=newOffsetIdx;
                  offsetIdx[newOffsetIdx+i+1]=item.offsetIdx;

                  if averageFlag:
                     CCD_speTmp[i]=CCD_Average(item);

               CCD_spe.append(CCD_speTmp[i]);

               if plotFlag:
                  CCD_Plot(CCD_speTmp[i]);

        nspectra=len(CCD_spe);
		
        if expected and nspectra and (nspectra<expected):
           for i in range(expected-nspectra):
               CCD_spe.append(CCD_spe[nspectra-1]);

    return CCD_spe

def CCD_SearchForDS(speFile):

    """
    CCD_SearchForDS(speFile)

    Search for direct sun with a exposure time less than 1 sec in the given file

    Parameters
    ----------

    speFile : the name of the spectra file to load                              \n

    """

    if not os.path.isfile(speFile):
        print('[CCD_SearchForDS] '+speFile+' not found !!!')
    else:

        f=open(speFile,"rb")

        nrec=0

        while True:

            # ----------------
            # READ NEXT RECORD
            # ----------------

            nrec=nrec+1
            spe=CCD_ReadNextRecord(f,nrec)
            if size(spe)==0: break;
            header=spe.header;

            if header['measureType']==CCD_MT_ds and header['exposureTime']<1 :

               print(sprintf('%-3d %02d:%02d:%02d, %.3f, %.2f',nrec,header['ti_hour'],header['ti_min'],header['ti_sec'],header['Zm'],header['exposureTime']))

        f.close()

#
# CCDVIS_SPW2DSC
# CCDUV_SPW2DSC
#
#  Update January 2016 :
#
#  * Improvements after comparison with Harestua Matlab routines
#
#    - add saturatedFlag in argument (otherwise if the acquisition included saturated scans, the spectrum was rejected)
#
#          saturatedFlag=0 : do not rejected spectra with saturated scans
#          saturatedFlag=1 : reject measurements including saturated scans
#
#    - if len(drk)<ntint : append last dark current measurement
#
#      By default, matlab program only loads 40 dark current spectra
#      Python routines load the 41 spectra if they exist (just that we check the number of spectra and append missing ones)
#
#    - use linear_least_squares instead of linalg.lstsq to solve \ matlab operator
#    - add straylight correction -> add range ???
#
#  * CCDUV_SPW2DSC : CINDI-2 bug fixed with last records : double loop on ccdY
#

def CCDVIS_SPW2DSC(spwFile,drkFile,straylightFile,varpixFile,trackFile,tintMin,dscFile,offsetFlag,saturatedFlag,useDrkOffset=0):

    """
    CCD_spe=CCDVIS_SPW2DSC(spwFile,drkFile,straylightFile,varpixFile,trackFile,tintMin,dscFile,offsetFlag,saturatedFlag,useDrkOffset=0)

    Save VIS spectra after some corrections (dark current, straylight, interpixel variability...)

    Parameters
    ----------

    spwFile       : name of the original spectra file (all types of measurements including offset)                   \n
    drkFile       : name of the dark current file                                                                    \n
    varpixFile    : name of the file with the interpixel variability matrix                                          \n
    trackFile     : track response of the instrument                                                                 \n
    tintMin       : included in order to have the same syntax as equivalent UV function but ignored for VIS spectra  \n
    dscFile       : the name of the output file (DSC extension recommended)                                          \n
    offsetFlag    : 0 : do not correct with offset                                                                   \n
                    1 : offset of the scan is used to correct spectra                                                \n
                    2 : use the closest offset measurement (schedule plan)                                           \n
    saturatedFlag : 1 to reject saturated spectra based on the saturated flag in header                              \n
    useDrkOffset  : 1 to use offset from the dark current instead of offset of the file

    Returns
    -------

    CCD_spe    : the spectra retrieved from the file

    See also
    --------

    CCD_spectrum,CCD_ReadNextRecord
    """

    tint=array([ 0.10,   0.20,   0.40,   0.50,   0.80,   1.00,   1.20,   1.50,   1.75,   2.00, \
                 2.50,   3.00,   3.60,   4.30,   5.20,   6.20,   7.50,   8.90,  11.00,  13.00, \
                16.00,  19.00,  22.00,  27.00,  32.00,  38.00,  46.00,  55.00,  66.00,  80.00, \
                95.00, 115.00, 140.00, 160.00, 200.00, 240.00, 280.00, 340.00, 410.00, 490.00, \
               590.00 ]);

    ntint=len(tint);                                                            # ntint=size(tint,1);
    A=array([tint,tint*tint,tint*tint*tint])                                    # A=[tint tint.*tint tint.*tint.*tint];
    drkFlag=0;

    # ----------------------------------
    # Load Stray-light Correction Matrix
    # ----------------------------------

    C=array([]);

    if len(straylightFile)>0:

       if not os.path.isfile(straylightFile):
          print('[CCDVIS_SPW2DSC] '+straylightFile+' not found !!!');
       else:

          print('Load straylightFile '+straylightFile);
          m=sio.loadmat(straylightFile);
          C=m['C'];

    # -----------------------------
    # Load VarPix Correction Matrix
    # -----------------------------

    # TODO

    # ----------------------------------------------
    # Load and pre-process Dark Current Measurements
    # ----------------------------------------------

    if len(drkFile)>0:

       if not os.path.isfile(drkFile):
          print('[CCDVIS_SPW2DSC] '+drkFile+' not found !!!');
       else:

          drkFlag=1;
          drk=CCD_ReadSpe(drkFile,1,0,0,0,ntint);                               # CCD_ReadDrkOff(drk_filename);

          ccdX=size(drk[0].spectrum,0);
          ccdY=size(drk[0].spectrum,1);                                         # [ccdX,ccdY]=size(drk(1).dark);

          # ---> Retrieve 3D Dark current Matrix (3rd order polyn. fit)
          # DRK [NPol, Npix, Ntrack];

          a=zeros((ntint,ccdY),dtype=double);
          DRK=zeros((size(A,0),ccdX,ccdY),dtype=float64);

          for npix in range(ccdX):                                              # for npix=1:ccdX
              for i in range(ntint):                                            #  for i=1:ntint
                  a[i,:]=array(drk[i].spectrumCorrected[npix,:])                #   a(i,:)=drk(i).darkCorrected(npix,:);
                                                                                #  end;
              for j in range(ccdY):                                             #  for j=1:ccdY
                  # p=linalg.lstsq(A.T,a[:,j])[0]                               #   p=A\a(:,j);
                  p=linear_least_squares(A.T,a[:,j]);
                  DRK[:,npix,j]=p;                                              #   DRK(:,npix,j)=p;
                                                                                #  end
                                                                                # end

    CCD_spe=[]                                                                  # array of spectra to return
    CCD_speTmp=[]                                                               # array of spectra not corrected yet

    lastOffsetIdx=-1;
    newOffsetIdx=-1;
    offsetIdx=[];

    if len(os.path.dirname(dscFile))==0:
       dscFile=os.path.curdir+'/'+dscFile

    if not os.path.isfile(spwFile):
        print('[CCD_SPW2DSC] '+spwFile+' not found !!!')
    elif not os.path.isdir(os.path.dirname(dscFile)):
        print('[CCD_SPW2DSC] '+os.path.dirname(dscFile)+' not a directory !!!')
    else:

        f=open(spwFile,"rb")
        g=open(dscFile,"wb")

        nrec=0

        # ---> BEGIN while True:

        while True:

            # ----------------
            # READ NEXT RECORD
            # ----------------

            nrec=nrec+1
            offsetIdx.append(-1);
            spe=CCD_ReadNextRecord(f,nrec)
            if size(spe)==0: break;
            header=spe.header;

            # ---> BEGIN if header['nacc'] # and header['nTint']:

            if header['nacc'] and ((saturatedFlag==0) or (header['saturatedFlag']==0)):    # and header['nTint']  comment for horizon measurements (no offset spectra, fixed integration time)

               # Measurements other than zenith and off-axis measurements : correct with the previous offset

               if (offsetFlag==0) or (offsetFlag==2) or (header['measureType']!=CCD_MT_zenith and header['measureType']!=CCD_MT_offaxis and header['measureType']!=CCD_MT_drk):

                  # New offset record

                  if header['measureType']!=CCD_MT_offset :
                     CCD_speTmp.append(spe);

                  elif useDrkOffset==0 :

                     if newOffsetIdx!=-1 :
                        lastOffset=newOffset;

                     lastOffsetIdx=newOffsetIdx;
                     newOffset=spe.spectrum;
                     newOffsetIdx=nrec;

                     for i,item in enumerate(CCD_speTmp):

                         # -----------------
                         # OFFSET CORRECTION
                         # -----------------

                         if (offsetFlag):

                            if ((offsetFlag==1) or (lastOffsetIdx==-1) or (nrec-(lastOffsetIdx+1)-i<=i+1) ) :
                               item.spectrumCorrected=item.spectrum-newOffset;
                               offsetIdx[lastOffsetIdx+i+1]=newOffsetIdx;
                            else:
                               item.spectrumCorrected=item.spectrum-lastOffset;
                               offsetIdx[lastOffsetIdx+i+1]=lastOffsetIdx;

                         else:
                            item.spectrumCorrected=item.spectrum;

                         if (drkFlag):

                             ti=item.header['exposureTime'];
                             A=array([ti,ti*ti,ti*ti*ti]);

                             for j in range(ccdY):                                  #  Apply dark current correction for each track in sequence

                                 # -----------------------------
                                 # APPLY DARK CURRENT CORRECTION
                                 # -----------------------------

                                 drk_current=dot(A,DRK[:,:,j]);                     #   drk_current=A*DRK(:,:,j);  % line vector of dark current for track j

                                 item.spectrumCorrected[:,j]=item.spectrumCorrected[:,j]-drk_current;

                         # -----------------
                         # VARPIX CORRECTION
                         # -----------------

                         # TODO

                         # ---------------
                         # AVERAGE SPECTRA
                         # ---------------

                         CCD_speTmp[i]=CCD_Average(item);

                         # ----------------------
                         # STRAY-LIGHT CORRECTION
                         # ----------------------

                         if ndim(C)==2 and size(C,0)>0 and size(C,1)>0 :
                            item.spectrumCorrected=dot(C,item.spectrumCorrected);

                         # -----------
                         # SAVE RECORD
                         # -----------

                         CCD_SaveSpe(g,CCD_speTmp[i]);
                         CCD_spe.append(CCD_speTmp[i]);

                     CCD_speTmp=[];

               # Zenith-sky and off-axis measurements : correct with the next offset measurement

               else:

                  CCD_speTmp.append(spe);

            # ---> END if header['nacc'] # and header['nTint']:

        # ---> END while True:

        # --------------------
        # PROCESS LAST RECORDS
        # --------------------

        if size(CCD_speTmp)>0:

           for i,item in enumerate(CCD_speTmp):

               # Offset correction

               if (drkFlag and useDrkOffset):

                  idrk=(abs(tint-item.header['exposureTime'])).argmin()
                  item.spectrumCorrected=item.spectrum-(drk[idrk].spectrum-drk[idrk].spectrumCorrected);

               elif useDrkOffset==0 and offsetFlag and 'newOffset' in locals():

                  item.spectrumCorrected=item.spectrum-newOffset;
                  offsetIdx[newOffsetIdx+i+1]=newOffsetIdx;

               else:

                  item.spectrumCorrected=item.spectrum;

               # Dark current correction

               if (drkFlag):

                  ti=item.header['exposureTime'];
                  A=array([ti,ti*ti,ti*ti*ti]);

                  for j in range(ccdY):

                      drk_current=dot(A,DRK[:,:,j]);
                      item.spectrumCorrected[:,j]=item.spectrumCorrected[:,j]-drk_current;

               # VarPix Correction

               # TODO

               # Average spectra

               CCD_speTmp[i]=CCD_Average(item);

               # Stray-light Correction

               if ndim(C)==2 and size(C,0)>0 and size(C,1)>0 :
                  item.spectrumCorrected=dot(C,item.spectrumCorrected);

               # Save record

               CCD_SaveSpe(g,item);
               CCD_spe.append(item);

        # Close files

        f.close()
        g.close()

    # plot(offsetIdx,'*');

    return CCD_spe;

def CCDUV_SPW2DSC(spwFile,drkFile,straylightFile,varpixFile,trackFile,tintMin,dscFile,offsetFlag,saturatedFlag,useDrkOffset=0):

    """
    CCD_spe=CCDUV_SPW2DSC(spwFile,drkFile,straylightFile,varpixFile,trackFile,tintMin,dscFile,offsetFlag,saturatedFlag,useDrkOffset=0)

    Save spectra after some corrections (dark current, straylight, interpixel variability...)

    Parameters
    ----------

    spwFile    : name of the original spectra file (all types of measurements including offset)   \n
    drkFile    : name of the dark current file                                                    \n
    varpixFile : name of the file with the interpixel variability matrix                          \n
    trackFile  : track response of the instrument                                                 \n
    tintMin    : integration time used to calculate the fitting polynomial                        \n
    dscFile    : the name of the output file (DSC extension recommended)                          \n
    offsetFlag : 0 : do not correct with offset                                                   \n
                 1 : offset of the scan is used to correct spectra                                \n
                 2 : use the closest offset measurement (schedule plan)                           \n
    saturatedFlag : 1 to reject saturated spectra based on the saturated flag in header           \n
    useDrkOffset  : 1 to use offset from the dark current instead of offset of the file           \n

    Returns
    -------

    CCD_spe    : the spectra retrieved from the file

    See also
    --------

    CCD_spectrum,CCD_ReadNextRecord
    """

    tint=array([ 0.10,   0.20,   0.40,   0.50,   0.80,   1.00,   1.20,   1.50,   1.75,   2.00, \
                 2.50,   3.00,   3.60,   4.30,   5.20,   6.20,   7.50,   8.90,  11.00,  13.00, \
                16.00,  19.00,  22.00,  27.00,  32.00,  38.00,  46.00,  55.00,  66.00,  80.00, \
                95.00, 115.00, 140.00, 160.00, 200.00, 240.00, 280.00, 340.00, 410.00, 490.00, \
               590.00 ]);

    ntint=len(tint);                                                            # ntint=size(tint,1);

    x=tint-tintMin;                                                             # x=tint-0.1;
    A=array([x,x*x,x*x*x,x*x*x*x]).T;                                           # A=[x x.*x x.*x.*x x.*x.*x.*x];
    x=tint-3.;                                                                  # x=tint-3
    B=array([x,x*x]).T;                                                         # B=[x x.*x];
    drkFlag=0;

    # ----------------------------------
    # Load Stray-light Correction Matrix
    # ----------------------------------

    C=array([]);

    if len(straylightFile)>0:

       if not os.path.isfile(straylightFile):
          print('[CCDUV_SPW2DSC] '+straylightFile+' not found !!!');
       else:

          print('Load straylightFile '+straylightFile);
          m=sio.loadmat(straylightFile);
          C=m['C'];

    # -----------------------------
    # Load VarPix Correction Matrix
    # -----------------------------

    # TODO

    # ----------------------------------------------
    # Load and pre-process Dark Current Measurements
    # ----------------------------------------------

    if len(drkFile)>0:

       if not os.path.isfile(drkFile):
          print('[CCDUV_SPW2DSC] '+drkFile+' not found !!!');
       else:

          print('Load drkFile '+drkFile);

          drk=CCD_ReadSpe(drkFile,1,0,0,0,ntint);                                     # CCD_ReadDrkOff(drk_filename);

          ccdX=size(drk[0].spectrum,0);
          ccdY=size(drk[0].spectrum,1);                                               # [ccdX,ccdY]=size(drk(1).dark);

          # ---> Retrieve 3D Dark current Matrix (3rd order polyn. fit)
          # DRK [NPol, Npix, Ntrack];

          a=zeros((ntint,ccdY),dtype=double);
          DRK=zeros((size(A,1)+size(B,1),ccdX,ccdY),dtype=float64);

          for npix in range(ccdX):                                                    # for npix=1:ccdX

              if (npix%100)==0:
                 print(sprintf('%d',npix))

              for i in range(ntint):                                                  #  for i=1:ntint
                  a[i,:]=array(drk[i].spectrum[npix,:])                               #   a(i,:)=drk(i).dark(npix,:); -> for UV, use not corrected dark current
                                                                                      #  end;
              for j in range(ccdY):                                                   #  for j=1:ccdY

                  # Tint < 3 sec

                  # p1=linalg.lstsq(A[0:12,:],a[0:12,j]-a[0,j])[0];                   #   p1=A(1:12,:)\(a(1:12,j)-a(1,j));
                  p1=linear_least_squares(A[0:12,:],a[0:12,j]-a[0,j]);
                  x=2.9;
                  dc=dot(array([x,x*x,x*x*x,x*x*x*x]),p1);                            #   dc=[x x*x x*x*x x*x*x*x]*p1;

                  # Tint > 3 sec
                  # p2=linalg.lstsq(B[11:ntint,0:2],(a[11:ntint,j]-a[0,j]-dc))[0];    #   p2=B(12:ntint,1:2)\(a(12:ntint,j)-a(1,j)-dc);
                  p2=linear_least_squares(B[11:ntint,0:2],a[11:ntint,j]-a[0,j]-dc);

                  DRK[0:4,npix,j]=p1;                                                 #   DRK(:,npix,j)=[p1;p2];
                  DRK[4:6,npix,j]=p2;
                                                                                      #  end
                                                                                      # end

    CCD_spe=[]                                                                  # array of spectra to return
    CCD_speTmp=[]                                                               # array of spectra not corrected yet

    lastOffsetIdx=-1;
    newOffsetIdx=-1;
    offsetIdx=[];

    if len(os.path.dirname(dscFile))==0:
       dscFile=os.path.curdir+'/'+dscFile

    if not os.path.isfile(spwFile):
        print('[CCDUV_SPW2DSC] '+spwFile+' not found !!!')
    elif not os.path.isdir(os.path.dirname(dscFile)):
        print('[CCDUV_SPW2DSC] '+os.path.dirname(dscFile)+' not a directory !!!')
    else:

        print('Load spwFile '+spwFile);

        f=open(spwFile,"rb")
        g=open(dscFile,"wb")

        nrec=0

        # ---> BEGIN while True:

        while True:

            # ----------------
            # READ NEXT RECORD
            # ----------------

            nrec=nrec+1
            offsetIdx.append(-1);

            print(sprintf('process file %s, spectrum no# %04d',spwFile,nrec));
            spe=CCD_ReadNextRecord(f,nrec)
            if size(spe)==0: break;
            header=spe.header;

            # ---> BEGIN if header['nacc'] and header['nTint']:

            if header['nacc'] and ((saturatedFlag==0) or (header['saturatedFlag']==0)):    # and header['nTint'] comment for horizon measurements (no offset spectra, fixed integration time)

               # Measurements other than zenith and off-axis measurements : correct with the previous offset

               if (offsetFlag==0) or (offsetFlag==2) or  (header['measureType']!=CCD_MT_zenith and header['measureType']!=CCD_MT_offaxis and header['measureType']!=CCD_MT_drk):

                  # New offset record

                  if header['measureType']!=CCD_MT_offset:
                     CCD_speTmp.append(spe);

                  elif useDrkOffset==0:

                     if newOffsetIdx!=-1 :
                        lastOffset=newOffset;

                     lastOffsetIdx=newOffsetIdx;
                     newOffset=spe.spectrum;
                     newOffsetIdx=nrec;

                     for i,item in enumerate(CCD_speTmp):

                         # -----------------
                         # OFFSET CORRECTION
                         # -----------------

                         if offsetFlag:

                            if ((offsetFlag==1) or (lastOffsetIdx==-1) or (nrec-(lastOffsetIdx+1)-i<=i+1) ) :
                               item.spectrumCorrected=item.spectrum-newOffset;
                               offsetIdx[lastOffsetIdx+i+1]=newOffsetIdx;
                            else:
                               item.spectrumCorrected=item.spectrum-lastOffset;
                               offsetIdx[lastOffsetIdx+i+1]=lastOffsetIdx;

                         else:
                            item.spectrumCorrected=item.spectrum;

                         # -----------------------------
                         # APPLY DARK CURRENT CORRECTION
                         # -----------------------------

                         if (drkFlag):

                            tint=item.header['exposureTime'];
                            x=tint-tintMin;                                        # x=tint-0.1;
                            A=array([x,x*x,x*x*x,x*x*x*x]).T;                      # A=[x x.*x x.*x.*x x.*x.*x.*x];
                            x=tint-3.;                                             # x=tint-3
                            B=array([x,x*x]).T;                                    # B=[x x.*x];

                            if tint<3.:

                               for j in range(ccdY):                               #  Apply dark current correction for each track in sequence

                                   drk_current=dot(A,DRK[0:4,:,j]);                #   drk_current=A*DRK(1:4,:,j);  % line vector of dark current for track j
                                   item.spectrumCorrected[:,j]=item.spectrumCorrected[:,j]-drk_current;

                            else:

                               for j in range(ccdY):                               #  Apply dark current correction for each track in sequence

                                   x=2.9;                                          # x=2.9; dc=[x, x*x, x*x*x, x*x*x*x]*DRK(1:4,:,jt)

                                   dc=dot(array([x,x*x,x*x*x,x*x*x*x]).T,DRK[0:4,:,j])

                                   drk_current=dot(B,DRK[4:6,:,j])+dc               # drk_current=B*DRK(5:6,:,jt)+dc;
                                   item.spectrumCorrected[:,j]=item.spectrumCorrected[:,j]-drk_current;

                         # -----------------
                         # VARPIX CORRECTION
                         # -----------------

                         # TODO

                         # ---------------
                         # AVERAGE SPECTRA
                         # ---------------

                         CCD_speTmp[i]=CCD_Average(item);

                         # ----------------------
                         # STRAY-LIGHT CORRECTION
                         # ----------------------

                         if ndim(C)==2 and size(C,0)>0 and size(C,1)>0 :

                            new=item.spectrumCorrected;
                            new2=dot(C,new);

                            sl=new-new2;

                            # Force linear behaviour on UV edge

                            x=arange(119.,300.);
                            y=sl[119:300,];
                            p=polyfit(x,y,1);

                            disp(p)

                            x=arange(0.,119.);

                            sl[0:119,]=polyval(p,x)-polyval(p,119)+sl[119,];

                            new2=new-sl;
                            item.spectrumCorrected=new2;

                         # -----------
                         # SAVE RECORD
                         # -----------

                         CCD_SaveSpe(g,CCD_speTmp[i]);
                         CCD_spe.append(CCD_speTmp[i]);

                     CCD_speTmp=[];

               # Zenith-sky and off-axis measurements : correct with the next offset measurement

               else:

                  CCD_speTmp.append(spe);

            # ---> END if header['nacc'] and header['nTint']:

        # ---> END while True:

        # --------------------
        # PROCESS LAST RECORDS
        # --------------------

        if size(CCD_speTmp)>0:

           for i,item in enumerate(CCD_speTmp):

               if (drkFlag and useDrkOffset):

                  idrk=(abs(tint-item.header['exposureTime'])).argmin()
                  item.spectrumCorrected=item.spectrum-(drk[idrk].spectrum-drk[idrk].spectrumCorrected);

               elif useDrkOffset==0 and offsetFlag>0 and 'newOffset' in locals():

                  item.spectrumCorrected=item.spectrum-newOffset;
                  offsetIdx[newOffsetIdx+i+1]=newOffsetIdx;

               # Dark current correction

               if (drkFlag):

                  tint=item.header['exposureTime'];
                  x=tint-tintMin;                                               # x=tint-0.1;
                  A=array([x,x*x,x*x*x,x*x*x*x]).T;                             # A=[x x.*x x.*x.*x x.*x.*x.*x];
                  x=tint-3.;                                                     # x=tint-3
                  B=array([x,x*x]).T;                                           # B=[x x.*x];

                  if tint<3.:

                     for j in range(ccdY):                                  #  Apply dark current correction for eauseDrkOffsetch track in sequence

                         drk_current=dot(A,DRK[0:4,:,j]);                   #   drk_current=A*DRK(1:4,:,j);  % line vector of dark current for track j
                         item.spectrumCorrected[:,j]=item.spectrumCorrected[:,j]-drk_current;

                  else:

                     for j in range(ccdY):                                  #  Apply dark current correction for each track in sequence

                         x=2.9;                                             # x=2.9; dc=[x, x*x, x*x*x, x*x*x*x]*DRK(1:4,:,jt)

                         dc=dot(array([x,x*x,x*x*x,x*x*x*x]).T,DRK[0:4,:,j])

                         drk_current=dot(B,DRK[4:6,:,j])+dc                 # drk_current=B*DRK(5:6,:,jt)+dc;
                         item.spectrumCorrected[:,j]=item.spectrumCorrected[:,j]-drk_current;

               # VarPix Correction

               # TODO

               # Average spectra

               CCD_speTmp[i]=CCD_Average(item);

               # Stray-light Correction

               if ndim(C)==2 and size(C,0)>0 and size(C,1)>0 :
                  item.spectrumCorrected=dot(C,item.spectrumCorrected);

               # Save record

               CCD_SaveSpe(g,item);
               CCD_spe.append(item);

        # Close files

        f.close()
        g.close()

    # plot(offsetIdx,'*');

    return CCD_spe;

def CCD_DSC2NC_WriteRecord(g,header,spe,k):

   # Get variables

   da_day=g.variables['da_day'];
   da_mon=g.variables['da_mon'];
   da_year=g.variables['da_year'];

   ti_hour=g.variables['ti_hour'];
   ti_min=g.variables['ti_min'];
   ti_sec=g.variables['ti_sec'];
   ti_msec=g.variables['ti_msec'];

   start_hour=g.variables['start_hour'];
   start_min=g.variables['start_min'];
   start_sec=g.variables['start_sec'];
   start_msec=g.variables['start_msec'];

   end_hour=g.variables['end_hour'];
   end_min=g.variables['end_min'];
   end_sec=g.variables['end_sec'];
   end_msec=g.variables['end_msec'];

   meas_type=g.variables['measurement_type'];
   view_el=g.variables['viewing_elevation_angle'];
   view_az=g.variables['viewing_azimuth_angle'];
   meas_time=g.variables['total_measurement_time'];
   acq_time=g.variables['total_acquisition_time'];
   tint=g.variables['exposure_time'];
   sza=g.variables['solar_zenith_angle'];
   saa=g.variables['solar_azimuth_angle'];

   spectrum=g.variables['spectrum'];

   measType=header['measureType'];

   if (measType==CCD_MT_offaxis) or (measType==CCD_MT_azimuth) or \
      (measType==CCD_MT_ds) or (measType==CCD_MT_almucantar) or \
      (measType==CCD_MT_zenith) or (measType==CCD_MT_horizon) :

      # Save header

      da_day[k]=header['da_day'];
      da_mon[k]=header['da_month'];
      da_year[k]=header['da_year'];

      ti_hour[k]=header['ti_hour'];
      ti_min[k]=header['ti_min'];
      ti_sec[k]=header['ti_sec'];
      ti_msec[k]=0;

      start_hour[k]=header['start_ti_hour'];
      start_min[k]=header['start_ti_min'];
      start_sec[k]=header['start_ti_sec'];
      start_msec[k]=0;

      end_hour[k]=header['end_ti_hour'];
      end_min[k]=header['end_ti_min'];
      end_sec[k]=header['end_ti_sec'];
      end_msec[k]=0;

      if (measType==CCD_MT_offaxis) or (measType==CCD_MT_azimuth) :
         meas_type[k]=CCD_MT_offaxis;
      elif (measType==CCD_MT_ds) or (measType==CCD_MT_almucantar) or (measType==CCD_MT_horizon) :
         meas_type[k]=measType;
      else:
         meas_type[k]=CCD_MT_zenith;

      view_el[k]=header['targetElevation'];
      view_az[k]=header['targetAzimuth'];
      meas_time[k]=(header['end_ti_hour']*3600+header['end_ti_min']*60+header['end_ti_sec']-   \
                   (header['start_ti_hour']*3600+header['start_ti_min']*60+header['start_ti_sec']));
      if (header['nTint']<=1):
         acq_time[k]=float(header['nacc']*header['exposureTime']);
      else:
         acq_time[k]=dot(header['nacc'],header['exposureTime']);

      tint[k]=header['exposureTime'];
      sza[k]=header['Zm'];
      saa[k]=header['Azimuth']+180.;

      # Save record

      spectrum[k,:]=reshape(spe,(1,len(spe)));

      k=k+1;

   # Return

   return k;

def CCD_DSC2NC(dscFile,hrzFile,ncFile,*args):

    """
    CCD_DSC2NC(dscFile,hrzFile,ncFile,[clbFile])

    Save spectra from DSC format to NETCDF format (CINDI-2)

    Parameters
    ----------

    dscFile    : name of the original spectra file                              \n
    hrzFile    : name of the horizon spectra file                               \n
    ncFile     : name of the output file (NETCDF format)                         \n
    clbFile    : optional, the calibration file (if given, will be saved in     \n
                 the output file)                                               \n

    """

    if not os.path.isfile(dscFile):
       print('[CCD_DSC2NC] '+dscFile+' not found !!!');
    elif not os.path.isdir(os.path.dirname(ncFile)):
       print('[CCD_DSC2NC] '+os.path.dirname(ncFile)+' not a directory !!!');
    else:

        dsc=CCD_ReadSpe(dscFile,0,0,0,0,0);
        hrz=CCD_ReadSpe(hrzFile,0,1,0,0,0);
        hrzInserted=0;

        nrec=size(dsc)+size(hrz);
        detector_size=(dsc[0].header['roiWveEnd']-dsc[0].header['roiWveStart']+1);

        if (size(dsc)>0):

           # Load calibration file

           if len(args)>0:

              clb=loadtxt(args[0],dtype=double);
              if ndim(clb)>1:
               clb=clb[:,0];
              clb=reshape(clb,(1,len(clb)));

           else:
              clb=[];

           # Open NetCDF file

           g=Dataset(ncFile,'w',format='NETCDF4_CLASSIC')

           # Write global attributes

           g.site="Cabauw";
           g.country="The Netherlands";
           g.latitude=51.97;
           g.longitude=4.93;
           g.altitude=0.;
           g.institute="bira";
           g.pi="Michel Van Roozendael (michel.vanroozendael@aeronomie.be)";
           g.data_creator="Caroline FAYT (caroline.fayt@aeronomie.be)";
           g.data_submitter="Caroline FAYT (caroline.fayt@aeronomie.be)";
           g.instrument_number=4;
           g.channel_number=2;
           g.detector_size=detector_size;
           g.number_of_records=nrec;
           g.calib=1 if len(clb)>0 else 0;

           # Create dimensions

           g.createDimension('number_of_records',nrec);
           g.createDimension('detector_size',detector_size);

           # Create groups

           # g.createGroup("header");
           # g.createGroup("date");
           # g.createGroup("time");
           # g.createGroup("start_time");
           # g.createGroup("end_time");
           # g.createGroup("data");

           # Create variables
           # complete with attributes such as units, range, fill values, description

           da_day=g.createVariable('da_day',int8,('number_of_records',),zlib=True);
           da_mon=g.createVariable('da_mon',int8,('number_of_records',),zlib=True);
           da_year=g.createVariable('da_year',int16,('number_of_records',),zlib=True);

           ti_hour=g.createVariable('ti_hour',int8,('number_of_records',),zlib=True);
           ti_min=g.createVariable('ti_min',int8,('number_of_records',),zlib=True);
           ti_sec=g.createVariable('ti_sec',int8,('number_of_records',),zlib=True);
           ti_msec=g.createVariable('ti_msec',int8,('number_of_records',),zlib=True);

           start_hour=g.createVariable('start_hour',int8,('number_of_records',),zlib=True);
           start_min=g.createVariable('start_min',int8,('number_of_records',),zlib=True);
           start_sec=g.createVariable('start_sec',int8,('number_of_records',),zlib=True);
           start_msec=g.createVariable('start_msec',int8,('number_of_records',),zlib=True);

           end_hour=g.createVariable('end_hour',int8,('number_of_records',),zlib=True);
           end_min=g.createVariable('end_min',int8,('number_of_records',),zlib=True);
           end_sec=g.createVariable('end_sec',int8,('number_of_records',),zlib=True);
           end_msec=g.createVariable('end_msec',int8,('number_of_records',),zlib=True);

           meas_type=g.createVariable('measurement_type',int8,('number_of_records',),zlib=True);
           view_el=g.createVariable('viewing_elevation_angle',float32,('number_of_records',),zlib=True);
           view_az=g.createVariable('viewing_azimuth_angle',float32,('number_of_records',),zlib=True);
           meas_time=g.createVariable('total_measurement_time',float32,('number_of_records',),zlib=True);
           acq_time=g.createVariable('total_acquisition_time',float32,('number_of_records',),zlib=True);
           tint=g.createVariable('exposure_time',float32,('number_of_records',),zlib=True);
           sza=g.createVariable('solar_zenith_angle',float32,('number_of_records',),zlib=True);
           saa=g.createVariable('solar_azimuth_angle',float32,('number_of_records',),zlib=True);

           spectrum=g.createVariable('spectrum',float64,('number_of_records','detector_size'),zlib=True);

           if (g.calib):
              calib=g.createVariable('calib',float64,('number_of_records','detector_size'),zlib=True);

           # Save calib

           if (g.calib):
              for i in arange(0,nrec):
                 calib[i,:]=clb;

           # Save spectra records

           k=0;
           for i in arange(0,size(dsc)):

              header=dsc[i].header;
              measType=header['measureType'];

              # Horizon measurements

              if not hrzInserted and size(hrz)>0 and ((header['ti_hour']+header['ti_min']/60.+header['ti_sec']/3600.)>(hrz[0].header['ti_hour']+hrz[0].header['ti_min']/60.+hrz[0].header['ti_sec']/3600.)):
                 for j in arange(0,size(hrz)):
                     k=CCD_DSC2NC_WriteRecord(g,hrz[j].header,hrz[j].spectrum,k)

                 hrzInserted=1;

              k=CCD_DSC2NC_WriteRecord(g,header,dsc[i].spectrum,k);

           g.horizon=hrzInserted;
           g.number_of_records=k;

           # Close file

           g.close()

def CCD_DSC2NC_Batch(clbFile,pathIn,pathHrz,pathOut,dateStart,dateEnd):

    date1=datetime.toordinal(datetime.strptime(dateStart,'%Y%m%d'))
    date2=datetime.toordinal(datetime.strptime(dateEnd,'%Y%m%d'))+1

    for dd in range(date1,date2):

        dateStr=datetime.strftime(datetime.fromordinal(dd),'%Y%m%d')
        dateYday=datetime.fromordinal(dd).timetuple().tm_yday

        fileIn=sprintf('%s/CBVIS_%s_%03d.DSC',pathIn,dateStr,dateYday);
        fileHrz=sprintf('%s/CBVIS_%s_%03d.hrz',pathHrz,dateStr,dateYday);
        fileOut=sprintf('%s/bira_maxdoas_4_vis_spectra_cindi2_%s_v1.asc',pathOut,dateStr);

        CCD_DSC2NC(fileIn,fileHrz,fileOut,clbFile);

def CCD_GetField(spe,fieldName):

    """
    fieldVector=CCD_GetField(spe,fieldName)

    retrieves the requested field from the header

    Parameters
    ----------

    spe : data returned by CCD_ReadSpe                                          \n
    fieldName : the requested field

    Returns
    -------

    fieldVector : the requested information                                     \n
    """

    nrec=len(spe);
    fieldVector=zeros(nrec,dtype=double);

    for i,item in enumerate(spe):

       fieldVector[i]=item.header[fieldName];

    # return

    return fieldVector;

def CCD_GetTime(ccdHeader):

    """
    (midUT,startUT,endUT)=CCD_GetFracTimes(ccdHeader)

    calculate fractional time for the current record (at middle, beginning and end of the measurement)

    Parameters
    ----------

    ccdHeader : header of the current record                                    \n

    Returns
    -------

    midUT   : fractional time at the middle of the measurement                  \n
    startUT : fractional time at the beginning of the measurement               \n
    endUT   : fractional time at the end of the measurement                     \n
    """

    midUT=datenum(ccdHeader['da_year'],ccdHeader['da_month'],ccdHeader['da_day'],ccdHeader['ti_hour'],ccdHeader['ti_min'],ccdHeader['ti_sec']);
    startUT=datenum(ccdHeader['da_year'],ccdHeader['da_month'],ccdHeader['da_day'],ccdHeader['start_ti_hour'],ccdHeader['start_ti_min'],ccdHeader['start_ti_sec']);
    endUT=datenum(ccdHeader['da_year'],ccdHeader['da_month'],ccdHeader['da_day'],ccdHeader['end_ti_hour'],ccdHeader['end_ti_min'],ccdHeader['end_ti_sec']);

    if (midUT<startUT): startUT=startUT-1;
    if (endUT<midUT): endUT=endUT+1;

    # return

    return(midUT,startUT,endUT);

def CCD_CalculateFluxes(speFile,clbFile,wve1,wve2,bandwidth,outFile):

    """
    CCD_CalculateFluxes(spwFile,clbFile,wve1,wve2,bandwidth,outFile)

    Calculate fluxes at two wavelengths

    Parameters
    ----------

    speFile    : name of the spectra file (SPW or DSC)                                            \n
    clbFile    : wavelength calibration (the size should be the same as the detector)             \n
    wve1       : first wavelength                                                                 \n
    wve2       : second wavelengh                                                                 \n
    bandwidth  : averaging bandwidth                                                              \n
    outFile    : the name of the output file                                                      \n
    """

    if not os.path.isfile(speFile):
       print('[CCD_CalculateFluxes] '+speFile+' not found !!!')
    elif not os.path.isfile(clbFile):
       print('[CCD_CalculateFluxes] '+clbFile+' not found !!!')
    elif not os.path.isdir(os.path.dirname(outFile)):
       print('[CCD_CalculateFluxes] '+os.path.dirname(outFile)+' not a directory !!!')
    else:

       # Load files

       clb=loadtxt(clbFile,dtype=double,usecols=arange(1))     # calibration
       data=CCD_ReadSpe(speFile,0,0,0,0,0);                    # spectra

       # Select ranges of indices around the given wavelengths

       delta=bandwidth*0.5

       n1=[i for (i,w) in enumerate(clb) if (w>=wve1-delta) and (w<=wve1+delta)]
       n2=[i for (i,w) in enumerate(clb) if (w>=wve2-delta) and (w<=wve2+delta)]

       if (len(n1)==0) or (len(n2)==0):

          print('[CCD_CalculateFluxes] Impossible to find the input wavelengths in the calibration file')

       else:

          g=open(outFile,"wb")

          for i,item in enumerate(data):

	            # Complete wavelengths

	            item.header['wve1']=wve1
	            item.header['wve2']=wve2

	            # Complete fluxes

	            item.header['flux1']=item.spectrum[n1].mean(0)
	            item.header['flux2']=item.spectrum[n2].mean(0)

	            # Save records

	            CCD_SaveSpe(g,item);

          g.close()

def CCD_CopyFluxes(fromFile,toFile,outFile):

    """
    CCD_CopyFluxes(fromFile,toFile,outFile)

    Copy fluxes from one file to another one, typically from the visible file to the UV one (NORS Xianghe)

    Parameters
    ----------

    fromFile   : file with fluxes to copy                   \n
    toFile     : file in which to copy fluxes               \n
    outFile    : output file (content of toFile with fluxes from fromFile)
    """

    if not os.path.isfile(fromFile):
       print('[CCD_CopyFluxes] '+fromFile+' not found !!!')
    elif not os.path.isfile(toFile):
       print('[CCD_CopyFluxes] '+toFile+' not found !!!')
    elif not os.path.isdir(os.path.dirname(outFile)):
       print('[CCD_CopyFluxes] '+os.path.dirname(outFile)+' not a directory !!!')
    else:

        # Load files

       fromData=CCD_ReadSpe(fromFile,0,0,0,0,0);                        # should be VIS spectra
       toData=CCD_ReadSpe(toFile,0,0,0,0,0);                            # should be UV spectra

       m=len(fromData)
       n=len(toData)

       g=open(outFile,"wb")

       j=0
       i=0

       iold=-1;     # index of the last record in fromFile with equivalence in toFile
       jold=-1;     # index of the last record in toFile with equivalence in fromFile

       while (True):

          copyFlag=0;

          if ((i==m) or (j==n)): break;

          fromHeader=fromData[i].header;
          toHeader=toData[j].header;

          (fromMidUT,fromStartUT,fromEndUT)=CCD_GetTime(fromHeader)
          (toMidUT,toStartUT,toEndUT)=CCD_GetTime(toHeader)

          # Overlapping

          if (((fromStartUT>=toStartUT) and (fromStartUT<=toEndUT)) or ((toStartUT>=fromStartUT) and (toStartUT<=fromEndUT))) :
             copyFlag=1;

          # The time of the last record in both files doesn't match -> suppose that it is correct

          elif ((i==m-1) and (j==n-1)):
             copyFlag=1;

          elif (toEndUT<fromStartUT):
             print(sprintf('No overlap j %d i %d %d-%d %f-%f %f-%f From %02d:%02d:%02d %02d:%02d:%02d End %02d:%02d:%02d %02d:%02d:%02d',j+1,i+1,fromData[i].header['measureType'],toData[j].header['measureType'],fromStartUT,fromEndUT,toStartUT,toEndUT,fromData[i].header['start_ti_hour'],fromData[i].header['start_ti_min'],fromData[i].header['start_ti_sec'],fromData[i].header['end_ti_hour'],fromData[i].header['end_ti_min'],fromData[i].header['end_ti_sec'],toData[j].header['start_ti_hour'],toData[j].header['start_ti_min'],toData[j].header['start_ti_sec'],toData[j].header['end_ti_hour'],toData[j].header['end_ti_min'],toData[j].header['end_ti_sec']))
             j=j+1;
          elif (fromEndUT<toStartUT):
             print(sprintf('No overlap i %d j %d %d-%d From %02d:%02d:%02d %02d:%02d:%02d End %02d:%02d:%02d %02d:%02d:%02d',i+1,j+1,fromData[i].header['measureType'],toData[j].header['measureType'],fromData[i].header['start_ti_hour'],fromData[i].header['start_ti_min'],fromData[i].header['start_ti_sec'],fromData[i].header['end_ti_hour'],fromData[i].header['end_ti_min'],fromData[i].header['end_ti_sec'],toData[j].header['start_ti_hour'],toData[j].header['start_ti_min'],toData[j].header['start_ti_sec'],toData[j].header['end_ti_hour'],toData[j].header['end_ti_min'],toData[j].header['end_ti_sec']))
             i=i+1;
          else:
             print(sprintf('Blocage ? i %d j %d From %02d:%02d:%02d %02d:%02d:%02d End %02d:%02d:%02d %02d:%02d:%02d',i+1,j+1,fromData[i].header['start_ti_hour'],fromData[i].header['start_ti_min'],fromData[i].header['start_ti_sec'],fromData[i].header['end_ti_hour'],fromData[i].header['end_ti_min'],fromData[i].header['end_ti_sec'],toData[j].header['start_ti_hour'],toData[j].header['start_ti_min'],toData[j].header['start_ti_sec'],toData[j].header['end_ti_hour'],toData[j].header['end_ti_min'],toData[j].header['end_ti_sec']))

          if copyFlag:

             # one record doesn't match, but preceeding one and succeeding one match -> assume that it also matches

             if (iold!=-1) and (jold!=-1) and (i-iold==j-jold):

                 while (jold<j-1):

                   iold=iold+1
                   jold=jold+1

                   # Complete wavelengths

                   toData[jold].header['wve1']=fromData[iold].header['wve1']
                   toData[jold].header['wve2']=fromData[iold].header['wve2']

                   # Complete fluxes

                   toData[jold].header['flux1']=fromData[iold].header['flux1']
                   toData[jold].header['flux2']=fromData[iold].header['flux2']

                   # Save record

                   print(sprintf('No overlap but save %d-%d %d-%d From %02d:%02d:%02d %02d:%02d:%02d End %02d:%02d:%02d %02d:%02d:%02d',iold+1,jold+1,fromData[iold].header['measureType'],toData[jold].header['measureType'],fromData[iold].header['start_ti_hour'],fromData[iold].header['start_ti_min'],fromData[iold].header['start_ti_sec'],fromData[iold].header['end_ti_hour'],fromData[iold].header['end_ti_min'],fromData[iold].header['end_ti_sec'],toData[jold].header['start_ti_hour'],toData[jold].header['start_ti_min'],toData[jold].header['start_ti_sec'],toData[jold].header['end_ti_hour'],toData[jold].header['end_ti_min'],toData[jold].header['end_ti_sec']));

                   CCD_SaveSpe(g,toData[jold]);

             elif (jold<j-1): # (jold!=-1) &

                 while (jold<j-1):

                   jold=jold+1
                   print(sprintf('No overlap save record %d anyway',jold+1))
                   CCD_SaveSpe(g,toData[jold]);

             # save the current record

             # Complete wavelengths

             toData[j].header['wve1']=fromData[i].header['wve1']
             toData[j].header['wve2']=fromData[i].header['wve2']

             # Complete fluxes

             toData[j].header['flux1']=fromData[i].header['flux1']
             toData[j].header['flux2']=fromData[i].header['flux2']

             # Save records

             CCD_SaveSpe(g,toData[j]);

             print(sprintf('Overlap %d-%d %d-%d %d-%d From %02d:%02d:%02d %02d:%02d:%02d End %02d:%02d:%02d %02d:%02d:%02d',i+1,j+1,iold+1,jold+1,fromData[i].header['measureType'],toData[j].header['measureType'],fromData[i].header['start_ti_hour'],fromData[i].header['start_ti_min'],fromData[i].header['start_ti_sec'],fromData[i].header['end_ti_hour'],fromData[i].header['end_ti_min'],fromData[i].header['end_ti_sec'],toData[j].header['start_ti_hour'],toData[j].header['start_ti_min'],toData[j].header['start_ti_sec'],toData[j].header['end_ti_hour'],toData[j].header['end_ti_min'],toData[j].header['end_ti_sec']));

             iold=i
             jold=j

             i=i+1
             j=j+1

       # missing records

       j=jold+1

       while (j<n):
          CCD_SaveSpe(g,toData[j]);
          j=j+1

       g.close()

       print(sprintf('[CCD_CopyFluxes] From %d/%d To %d/%d',i,m,j,n))

def CCD_CompareFiles(file1,file2):

    """
    CCD_CompareFiles(file1,file2)

    Compare files (size, date and time) after update of fluxes

    Parameters
    ----------

    file1      : original file                  \n
    file2      : file after update              \n
    """

    if not os.path.isfile(file1):
       print('[CCD_CompareFiles] '+file1+' not found !!!')
    elif not os.path.isfile(file2):
       print('[CCD_CompareFiles] '+file2+' not found !!!')
    else:

       fromData=CCD_ReadSpe(file1,0,0,0,0,0);                          # should be original file
       toData=CCD_ReadSpe(file2,0,0,0,0,0);                            # should be file after update (for example, fluxes)

       m=len(fromData)
       n=len(toData)

       if (m!=n):
          print('[CCD_CompareFiles] Files do not have the same size')
       else:

          i=0

          for i in arange(m):

              fromHeader=fromData[i].header;
              toHeader=toData[i].header;

              if (fromHeader['ti_hour']!=toHeader['ti_hour']) or (fromHeader['ti_min']!=toHeader['ti_min']) or (fromHeader['ti_sec']!=toHeader['ti_sec']):
                 break;

          if (m>0) and (i!=m-1):
             print(sprintf('[CCD_CompareFiles] Files have different records (%d)',i))
          else:
             print(sprintf('%s : OK',os.path.basename(file1)))

def CCD_CheckAngles(station,file,figNumber):

    """
    CCD_CheckAzimuthAngles(station,file,figNumber)

    Check azimuth angles (useful for Southern hemisphere)

    Parameters
    ----------

    station : the name of the station                                           \n
    file : the name of the file to check                                        \n
    figNumber : the figure number
    """

    if not os.path.isfile(file):
       print('[CCD_CheckAzimuthAngles] '+file+' not found !!!')
    else:

       str=os.path.basename(file);
       a=str.split('_')
       dateStr=a[1];

       data=CCD_ReadSpe(file,0,1,0,0,0);

       hh=CCD_GetField(data,'ti_hour');
       mm=CCD_GetField(data,'ti_min');
       ss=CCD_GetField(data,'ti_sec');

       ut=hh+mm/60+ss/3600;

       targetElev=CCD_GetField(data,'targetElevation');
       trackerElev=CCD_GetField(data,'brusagElevation');
       targetAzim=CCD_GetField(data,'targetAzimuth');
       trackerAzim=CCD_GetField(data,'brusagAzimuth');
       tint=CCD_GetField(data,'exposureTime');
       measType=CCD_GetField(data,'measureType');
       ds=[i for (i,m) in enumerate(measType) if (m==CCD_MT_ds)]       # find direct sun measurements only

       figure(figNumber);
       clf();

       subplot(3,1,1);
       plot(ut,targetElev,'*',ut,trackerElev,'.r')
       title(sprintf('Elevation viewing angle at %s - %s/%s/%s',station,dateStr[6:8],dateStr[4:6],dateStr[0:4]),fontweight='bold',fontsize=14);
       legend(('Target','Tracker'));
       grid();
       axis([2,18,0,100])

       subplot(3,1,2);
       plot(ut,targetAzim,'*',ut,trackerAzim,'.r')
       title(sprintf('Azimuth viewing angle at %s - %s/%s/%s',station,dateStr[6:8],dateStr[4:6],dateStr[0:4]),fontweight='bold',fontsize=14);
       legend(('Target','Tracker'));
       grid();
       axis([2,18,-200,200])

       subplot(3,1,3);
       plot(ut,tint,'.',ut[ds],tint[ds],'.r')
       title(sprintf('Integration time at %s - %s/%s/%s',station,dateStr[6:8],dateStr[4:6],dateStr[0:4]),fontweight='bold',fontsize=14);
       legend(('all','direct sun'));
       grid();
       axis([2,18,0,3])

def CCD_RetrieveFluxes(dataSpe,dataInfo,clbFile,wve,bandwidth):

    if not os.path.isfile(clbFile):
       print('[CCD_RetrieveFluxes] '+clbFile+' not found !!!')
    else:

       # Load files

       clb=loadtxt(clbFile,dtype=double,usecols=arange(1))     # calibration

       # Select ranges of indices around the given wavelengths

       delta=bandwidth*0.5

       n1=[i for (i,w) in enumerate(clb) if ((w>=wve-delta) and (w<=wve+delta))];

       if (len(n1)==0):
          print('[CCD_RetrieveFluxes] Impossible to find the input wavelength in the calibration file');

       # Complete fluxes

       else:

          for i,item in enumerate(dataSpe):

             dataInfo.flux[i]=item.spectrum[n1].mean(0)/dataInfo.tint[i];

def CCD_InvertSpectrum(inputFile,outputFile):
    """
    Invert spectra of inputFile; result is written to outputFile

    Parameters
    ----------
                                                                                \n
    inputFile  : File with spectra to invert                                    \n
    outputFile : File with inverted spectra                                     \n
    """

    spe=[]

    if not os.path.isfile(inputFile):
        print('[CCD_InvertSpectrum] '+inputFile+' not found !!!')
    elif not os.path.isdir(os.path.dirname(outputFile)):
        print('[CCD_InvertSpectrum] '+os.path.dirname(outputFile)+' not a directory !!!')
    else:

        try:

           g=open(outputFile,"wb");
           spe=CCD_ReadSpe(inputFile,0,0,0,0,0);

           for i in arange(0,len(spe)):

              if spe[i].header['nacc'] > 0:

                # if (i>=581):                      # For 27/08/2018 only
                #   new_spectrum=spe[i].spectrum;
                # else:
                new_spectrum=flip(spe[i].spectrum,0);

                if spe[i].header['nTint']:
                  new_spe=CCD_spectrum(i,spe[i].header,new_spectrum,new_spectrum,spe[i].tint,spe[i].ntint,spe[i].specMax);
                else:
                  new_spe=CCD_spectrum(i,spe[i].header,new_spectrum,new_spectrum);

                CCD_SaveSpe(g,new_spe)

           g.close();

           # for i in arange(0,len(spe):
           # CCD_ReadSpe(speFile,offsetFlag,averageFlag,plotFlag,measType,expected):

        except BaseException as error:
           print('[CCD_InvertSpectrum] returned fatal error : {}'.format(error))

    return spe;

def CCD_RetrieveFileInfo(data):
    """
    Retrieve all useful information from the spectrum file

    Parameters
    ----------
                                                                                \n
    data  : data returned by CCD_ReadSpe                                        \n
    """

    ut=CCD_GetField(data,'ti_hour')+CCD_GetField(data,'ti_min')/60+CCD_GetField(data,'ti_sec')/3600;
    measType=CCD_GetField(data,'measureType');
    tint=CCD_GetField(data,'exposureTime');
    totalTime=CCD_GetField(data,'totalTime');
    sza=CCD_GetField(data,'Zm');
    sazim=CCD_GetField(data,'Azimuth');
    nrej=CCD_GetField(data,'nrej');
    nscans=CCD_GetField(data,'nacc');
    ntint=CCD_GetField(data,'nTint');
    trackerElevation=CCD_GetField(data,'brusagElevation');
    trackerAzimuth=CCD_GetField(data,'brusagAzimuth');
    filter=CCD_GetField(data,'filterNumber');
    targetElevation=CCD_GetField(data,'targetElevation');
    targetAzimuth=CCD_GetField(data,'targetAzimuth');
    startTime=CCD_GetField(data,'start_ti_hour')+CCD_GetField(data,'start_ti_min')/60+CCD_GetField(data,'start_ti_sec')/3600;
    endTime=CCD_GetField(data,'end_ti_hour')+CCD_GetField(data,'end_ti_min')/60+CCD_GetField(data,'end_ti_sec')/3600;
    jewell=CCD_GetField(data,'jewell');
    saturatedFlag=CCD_GetField(data,'saturatedFlag');

    ndata=len(ut);
    flux=zeros(ndata);

    # for i,item in enumerate(data):
    #    if (tint[i]>0.):
    #       flux[i]=max(item.spectrum)/tint[i];

    return CCD_info(ut,measType,tint,totalTime,sza,sazim,nrej,nscans,ntint,trackerElevation,trackerAzimuth,filter,targetElevation,targetAzimuth,startTime,endTime,jewell,flux,saturatedFlag);

def CCD_GetSpectrum(data,indexColumn):

   spe=[];
   n=len(indexColumn);

   if n>0:

      spe=zeros([size(data[0].spectrum),n]);
      for i in arange(0,n):
         spe[:,i]=data[indexColumn[i]].spectrum;

   return spe;

def CCD_MergeFiles(in1,in2,out)   :

    if not os.path.isfile(in1):
       print('[CCD_MergeFiles] '+fileName+' not found !!!')
    elif not os.path.isfile(in2):
       print('[CCD_MergeFiles] '+fileName+' not found !!!')
    elif not os.path.isdir(os.path.dirname(out)):
        print('[out] '+os.path.dirname(out)+' not a directory !!!')
    else:

       g=open(out,"wb")

       spe1=CCD_ReadSpe(in1,0,0,0,0,0);
       spe2=CCD_ReadSpe(in2,0,0,0,0,0);

       nspe1=len(spe1);
       nspe2=len(spe2);

       ispe1=0;
       ispe2=0;

       while (ispe1<nspe1) or (ispe2<nspe2):

           if (ispe1<nspe1) and (ispe2<nspe2):

             (midUT1,startUT1,endUT1)=CCD_GetTime(spe1[ispe1].header)
             (midUT2,startUT2,endUT2)=CCD_GetTime(spe2[ispe2].header)

             if midUT1<midUT2:
                CCD_SaveSpe(g,spe1[ispe1]);
                ispe1=ispe1+1;
             elif midUT2<midUT1:
                CCD_SaveSpe(g,spe2[ispe2]);
                ispe2=ispe2+1;
             elif midUT2==midUT1:
                CCD_SaveSpe(g,spe1[ispe1]);
                ispe1=ispe1+1;
                CCD_SaveSpe(g,spe2[ispe2]);
                ispe2=ispe2+1;

           elif (ispe1<nspe1):
                CCD_SaveSpe(g,spe1[ispe1]);
                ispe1=ispe1+1;
           else:
                CCD_SaveSpe(g,spe2[ispe2]);
                ispe2=ispe2+1;

       g.close();

# Reference spectra are in ASCII (save them to CCD_SPE structure using an example)

def CCD_GetRef_asc2spw(fileName,dd,nacc,tint,elevation,azimuth,spe_example):

   ref=loadtxt(fileName,comments=';');                                          # load the file
   spe=spe_example;                                                             # get a copy of the example (the number of integration time should be 1)
   spe.spectrum=ref;                                                            # replace spectrum by the reference
   spe.spectrumCorrected=ref;                                                   # replace corrected spectrum by the reference (no correction applied)

   spe.specMax=30000*ones(nacc);

   dd_delta=timedelta(seconds=int(nacc*tint*0.5));                              # delta time to calculate the
   dd_start=dd-dd_delta
   dd_end=dd+dd_delta

   # date and time fields

   spe.header['ti_hour']=dd.hour;
   spe.header['ti_min']=dd.minute;
   spe.header['ti_sec']=dd.second;

   spe.header['start_ti_hour']=dd_start.hour;
   spe.header['start_ti_min']=dd_start.minute;
   spe.header['start_ti_sec']=dd_start.second;

   spe.header['end_ti_hour']=dd_end.hour;
   spe.header['end_ti_min']=dd_end.minute;
   spe.header['end_ti_sec']=dd_end.second;

   spe.header['nacc']=nacc;                                                     # number of accumulations
   spe.tint=array([tint]);                                                      # exposure time (only one integration time)
   spe.header['exposureTime']=tint                                              # exposure time
   spe.header['totalTime']=int32(tint*nacc);                                    # total time
   spe.header['targetAzimuth']=azimuth;
   spe.header['targetElevation']=elevation;
   spe.header['brusagAzimuth']=azimuth;
   spe.header['brusagElevation']=elevation;

   (sza,azim)=sunangl(array([dd.year,dd.month,dd.day]),array([[dd.hour,dd.minute,dd.second]]),-4.36,50);

   spe.header['Zm']=sza;

   return spe

def CCD_Horizon(fileName,clbFile,wve1,wve2):

    if not os.path.isfile(fileName):
       print('[CCD_Horizon] '+fileName+' not found !!!')
    elif not os.path.isfile(clbFile):
       print('[CCD_Horizon] '+clbFile+' not found !!!')
    else:

       hrz=CCD_ReadSpe(fileName,0,1,0,0,0);
       hrz_tint=CCD_GetField(hrz,"exposureTime");
       hrz_elev=CCD_GetField(hrz,"targetElevation");

       clb=loadtxt(clbFile,comments=';');
       i=[j for (j,item) in enumerate(clb[:,0]) if (item>=wve1 and item<=wve2)];

       hrz_flux=zeros(size(hrz_tint));

       for j in arange(0,size(hrz_elev)):
          hrz_flux[j]=max(hrz[j].spectrum[i])/hrz_tint[j];

       return (hrz_elev,hrz_flux)



